# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from skimage.color import hsv2rgb

import numpy as np
import SimpleITK as sitk

import matplotlib.pyplot as plt

from platipy.imaging.utils.crop import label_to_roi


class VisualiseImage:
    """Class to represent visualisation of an image"""

    def __init__(self, image, aspect, interpolation, origin, colormap, clim):
        self.image = image
        self.aspect = aspect
        self.interpolation = interpolation
        self.origin = origin
        self.colormap = colormap
        self.clim = clim


class VisualiseContour:
    """Class to represent the visualiation of a contour"""

    def __init__(self, image, name, color=None, linewidth=2, linestyle="solid"):
        self.image = image
        self.name = name
        self.color = color
        self.linewidth = linewidth
        self.linestyle = linestyle


class VisualiseScalarOverlay:
    """Class to represent the visualiation of a scalar overlay"""

    def __init__(
        self,
        image,
        name,
        colormap=plt.cm.get_cmap("Spectral"),
        alpha=0.75,
        min_value=False,
        max_value=False,
        discrete_levels=False,
        mid_ticks=False,
        show_colorbar=True,
        norm=None,
        projection=False,
    ):
        self.image = image
        self.name = name
        self.colormap = colormap
        self.alpha = alpha
        self.min_value = min_value
        self.max_value = max_value
        self.discrete_levels = discrete_levels
        self.mid_ticks = mid_ticks
        self.show_colorbar = show_colorbar
        self.norm = norm
        self.projection = projection


class VisualiseVectorOverlay:
    """Class to represent the visualiation of a vector overlay"""

    def __init__(
        self,
        image,
        min_value=False,
        max_value=False,
        colormap=plt.cm.get_cmap("Spectral"),
        discrete_levels=False,
        mid_ticks=False,
        alpha=0.75,
        arrow_scale=0.25,
        arrow_width=1,
        subsample=4,
        color_function="perpendicular",
        invert_field=True,
        show_colorbar=True,
        name="input",
    ):
        self.image = image
        self.min_value = min_value
        self.max_value = max_value
        self.colormap = colormap
        self.discrete_levels = discrete_levels
        self.mid_ticks = mid_ticks
        self.alpha = alpha
        self.arrow_scale = arrow_scale
        self.arrow_width = arrow_width
        self.subsample = subsample
        self.color_function = color_function
        self.invert_field = invert_field
        self.show_colorbar = show_colorbar
        self.name = name


class VisualiseComparisonOverlay:
    """Class to represent the visualiation of a comparison image"""

    def __init__(self, image, name, color_rotation=0.35):
        self.image = image
        self.name = name
        self.color_rotation = color_rotation


class VisualiseBoundingBox:
    """Class to represent the visualiation of a bounding box"""

    def __init__(self, bounding_box, name, color="r", linewidth=2):

        if isinstance(bounding_box, sitk.Image):
            bounding_box = label_to_roi(bounding_box, return_as_list=True)

        self.bounding_box = bounding_box
        self.name = name
        self.color = color
        self.linewidth = linewidth


def return_slice(axis, index):
    """Prepares a slice tuple to use for extracting a slice for rendering

    Args:
        axis (str): One of "x", "y" or "z"
        index (int): The index of the slice to fetch

    Returns:
        tuple: can be used to extract a slice
    """

    if axis == "x":
        return (slice(None), slice(None), index)
    if axis == "y":
        return (slice(None), index, slice(None))
    if axis == "z":
        return (index, slice(None), slice(None))

    return None


def subsample_vector_field(axis, cut, subsample=1):
    """Prepares a slice tuple to use for extracting a slice for rendering

    Args:
        axis (str): One of "x", "y" or "z"
        cut (int): The index of the image slice
        subsample (int): the subsample factor

    Returns:
        tuple: can be used to extract a vector field slice
    """
    if hasattr(subsample, "__iter__"):
        subsample_ax, subsample_cor, subsample_sag = subsample
    else:
        subsample_ax, subsample_cor, subsample_sag = (subsample,) * 3

    if axis == "x":
        return (
            slice(None, None, subsample_ax),
            slice(None, None, subsample_cor),
            cut,
        )
    if axis == "y":
        return (
            slice(None, None, subsample_ax),
            cut,
            slice(None, None, subsample_sag),
        )
    if axis == "z":
        return (
            cut,
            slice(None, None, subsample_cor),
            slice(None, None, subsample_sag),
        )
    return None


def vector_image_grid(axis, vector_field_array, subsample=1):
    """Prepares a grid for rendering a vector field on an image

    Args:
        axis (str): One of "x", "y" or "z"
        vector_field_array (np.array): the vector field array
        subsample (int): the subsample factor

    Returns:
        tuple: defines the 2 dimensional grid for displaying vectors
    """
    if hasattr(subsample, "__iter__"):
        subsample_ax, subsample_cor, subsample_sag = subsample
    else:
        subsample_ax, subsample_cor, subsample_sag = (subsample,) * 3

    if axis == "x":
        return np.mgrid[
            0 : vector_field_array.shape[1] : subsample_cor,
            0 : vector_field_array.shape[0] : subsample_ax,
        ]
    if axis == "y":
        return np.mgrid[
            0 : vector_field_array.shape[2] : subsample_sag,
            0 : vector_field_array.shape[0] : subsample_ax,
        ]
    if axis == "z":
        return np.mgrid[
            0 : vector_field_array.shape[2] : subsample_sag,
            0 : vector_field_array.shape[1] : subsample_cor,
        ]
    return None


def reorientate_vector_field(axis, vector_ax, vector_cor, vector_sag, invert_field=True):
    """Reorients vector field components for rendering
    This is necessary after converting from sitk.Image to np.array

    Args:
        axis (str): One of "x", "y" or "z"
        vector_ax (np.array): The first vector component (z)
        vector_cor (np.array): The second vector component (y)
        vector_sag (np.array): The third vector component (x)

    Returns:
        tuple: the re-oriented vector field components
    """

    if invert_field:
        vector_ax = -vector_ax
        vector_cor = -vector_cor
        vector_sag = -vector_sag

    if axis == "x":  # sagittal projection
        return vector_cor, vector_ax, vector_sag
    if axis == "y":  # coronal projection
        return vector_sag, vector_ax, vector_cor
    if axis == "z":  # axial projection
        return vector_sag, -vector_cor, vector_ax

    return None


def generate_comparison_colormix(
    image_list, arr_slice=None, window=(-250, 500), color_rotation=0.35
):

    #! TO DO - make this function take in more than two images
    # Will need to use polar coordinates for HSV colorspace addition

    if len(image_list) == 2:
        if all([isinstance(image, sitk.Image) for image in image_list]):
            if any([image.GetDimension() >= 3 for image in image_list]) and arr_slice is None:
                raise ValueError("Images cannot be 3D unless 'arr_slice' is specified.")

            else:
                array_list = [
                    sitk.GetArrayViewFromImage(image).__getitem__(arr_slice)
                    for image in image_list
                ]

        elif all([isinstance(image, np.ndarray) for image in image_list]):
            if any([len(image.shape) >= 3 for image in image_list]) and arr_slice is None:
                raise ValueError("Images cannot be 3D unless 'arr_slice' is specified.")

            else:
                array_list = [image.__getitem__(arr_slice) for image in image_list]

    else:
        raise ValueError("'image_list' must be a list of two sitk.Image or np.ndarray.")

    nda_a, nda_b = array_list

    nda_a_norm = (np.clip(nda_a, window[0], window[0] + window[1]) - window[0]) / (window[1])
    nda_b_norm = (np.clip(nda_b, window[0], window[0] + window[1]) - window[0]) / (window[1])

    nda_color = np.stack(
        [
            color_rotation * (nda_a_norm > nda_b_norm)
            + (0.5 + color_rotation) * (nda_a_norm <= nda_b_norm),
            np.abs(nda_a_norm - nda_b_norm),
            (nda_a_norm + nda_b_norm) / 2,
        ],
        axis=-1,
    )

    return hsv2rgb(nda_color)


def project_onto_arbitrary_plane(
    image,
    projection_name="mean",
    projection_axis=0,
    rotation_axis=[1, 0, 0],
    rotation_angle=0,
    default_value=-1000,
    resample_interpolation=2,
):
    """[summary]

    Args:
        image ([type]): [description]
        projection_name (str, optional): [description]. Defaults to "mean".
        projection_axis (int, optional): [description]. Defaults to 0.
        rotation_axis (list, optional): [description]. Defaults to [1, 0, 0].
        rotation_angle (int, optional): [description]. Defaults to 0.
        default_value (int, optional): [description]. Defaults to -1000.
        resample_interpolation (int, optional): [description]. Defaults to 2.

    Returns:
        [type]: [description]
    """

    projection_dict = {
        "sum": sitk.SumProjection,
        "mean": sitk.MeanProjection,
        "median": sitk.MedianProjection,
        "std": sitk.StandardDeviationProjection,
        "min": sitk.MinimumProjection,
        "max": sitk.MaximumProjection,
    }
    projection_function = projection_dict[projection_name]

    # Set centre as image centre
    rotation_centre = image.TransformContinuousIndexToPhysicalPoint(
        [(index - 1) / 2.0 for index in image.GetSize()]
    )

    # Define the transform, using predefined centre of rotation and given angle
    rotation_transform = sitk.VersorRigid3DTransform()
    rotation_transform.SetCenter(rotation_centre)
    rotation_transform.SetRotation(rotation_axis, rotation_angle)

    # Resample the image using the rotation transform
    resampled_image = sitk.Resample(
        image,
        rotation_transform,
        resample_interpolation,
        default_value,
        image.GetPixelID(),
    )

    # Project onto the given axis
    proj_image = projection_function(resampled_image, projection_axis)

    # Return this view
    image_slice = {
        0: proj_image[0, :, :],
        1: proj_image[:, 0, :],
        2: proj_image[:, :, 0],
    }

    return image_slice[projection_axis]



# Copyright 2022 Radiotherapy AI Pty Ltd

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pathlib
import shutil
import tempfile
import urllib.request

import matplotlib.pyplot as plt
import SimpleITK as sitk

from platipy.dicom.io.rtstruct_to_nifti import convert_rtstruct
from platipy.dicom.io.rtstruct_to_nifti import read_dicom_image
from platipy.imaging import ImageVisualiser


def teardown_function():
    plt.close()


def test_image_visualiser_use_case():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = pathlib.Path(temp_dir)

        zip_url = (
            "https://github.com/RadiotherapyAI/test-data-public/"
            "releases/download/deepmind-dicom/0522c0768.zip"
        )
        zip_filepath = temp_dir / "data.zip"

        urllib.request.urlretrieve(zip_url, zip_filepath)
        shutil.unpack_archive(zip_filepath, temp_dir)

        patient_dir = temp_dir / "0522c0768"
        structure_path = patient_dir / "RS.dcm"

        ct_images = read_dicom_image(patient_dir)
        prefix = "structure_masks_"

        convert_rtstruct(
            patient_dir,
            structure_path,
            prefix=prefix,
            output_dir=patient_dir,
            spacing=None,
        )

        structure_paths = patient_dir.glob(f"{prefix}*.nii.gz")

        contours = {}
        for path in structure_paths:
            filename = _name_with_all_suffixes_removed(path)
            contour_name = removeprefix(filename, prefix)

            contours[contour_name] = sitk.ReadImage(str(path))


    image_visualiser = ImageVisualiser(ct_images)
    image_visualiser.add_contour(contours, color="red", linewidth=1)
    fig = image_visualiser.show()

    assert fig is not None

    return fig


def _name_with_all_suffixes_removed(path: pathlib.Path):
    while path.suffix:
        path = path.with_suffix("")

    return path.name


# https://peps.python.org/pep-0616/#specification
def removeprefix(string: str, prefix: str) -> str:
    if string.startswith(prefix):
        return string[len(prefix):]
    else:
        return string[:]



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
import os
import uuid

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_restful import Api
from celery import Celery
import celery.signals

from platipy.backend.application import FlaskApp

env_work = os.getcwd()
if "WORK" in os.environ:
    env_work = os.environ["WORK"]
log_file_path = os.path.join(env_work, "service.log")


def configure_logging():
    logger = logging.getLogger()

    logger.handlers.clear()
    logger.setLevel(logging.DEBUG)

    file_handler = logging.handlers.RotatingFileHandler(
        log_file_path,
        maxBytes=100 * 1024 * 1024,  # Max 100 MB per log file before rotating
        backupCount=100,  # Keep up to 100 log files in history
    )
    file_formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    file_handler.setFormatter(file_formatter)
    file_handler.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)

    console_handler = logging.StreamHandler(sys.stdout)
    console_formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    console_handler.setFormatter(console_formatter)
    console_handler.setLevel(logging.DEBUG)
    logger.addHandler(console_handler)


@celery.signals.setup_logging.connect
def on_celery_setup_logging(**kwargs):
    configure_logging()


configure_logging()

# Create Flask app
app = FlaskApp(__name__)
app.config["SECRET_KEY"] = uuid.uuid4().hex

# Configure SQL Alchemy
app.config["SQLALCHEMY_DATABASE_URI"] = f"sqlite:///{env_work}/{os.path.basename(os.getcwd())}.db"
app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
db = SQLAlchemy(app)

# Configure Celery
REDIS_HOST = "localhost"
if "REDIS_HOST" in os.environ:
    REDIS_HOST = os.environ["REDIS_HOST"]

REDIS_PORT = 6379
if "REDIS_PORT" in os.environ:
    REDIS_PORT = os.environ["REDIS_PORT"]

app.config["CELERY_BROKER_URL"] = f"redis://{REDIS_HOST}:{REDIS_PORT}/0"
app.config["result_backend"] = app.config["CELERY_BROKER_URL"]
app.config["worker_max_memory_per_child"] = 32 * 1000 * 1000  # 16GB in KB
celery = Celery(__name__, broker=app.config["CELERY_BROKER_URL"])
celery.conf.update(app.config)

# Configure API
api = Api(app)
app.config.from_object("platipy.backend.api.CustomConfig")

import platipy.backend.views
import platipy.backend.api
import platipy.backend.tasks
from platipy.backend.models import DataObject



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=redefined-outer-name

import tempfile
from pathlib import Path

import pytest
import pydicom
import numpy as np
import SimpleITK as sitk

from platipy.imaging.tests.data import get_lung_dicom, get_lung_nifti
from platipy.dicom.io.rtstruct_to_nifti import convert_rtstruct
from platipy.dicom.io.nifti_to_rtstruct import convert_nifti


@pytest.fixture
def dicom_data():
    """Gets the data needed for converting DICOM

    Returns:
        pathlib.Path: Path to DICOM data
    """
    return get_lung_dicom()


@pytest.fixture
def nifti_data():
    """Gets the data needed for converting NIFTI

    Returns:
        pathlib.Path: Path to NIFTI data
    """
    return get_lung_nifti()


def test_convert_dicom_to_nifti(dicom_data):

    test_pat_id = "LCTSC-Test-S1-101"
    ct_uid = "1.3.6.1.4.1.14519.5.2.1.7014.4598.106943890850011666503487579262"
    rtstruct_uid = "1.3.6.1.4.1.14519.5.2.1.7014.4598.280355341349691222365783556597"

    pat_path = dicom_data.joinpath(test_pat_id)
    ct_path = pat_path.joinpath(ct_uid)
    rtstruct_path = next(pat_path.joinpath(rtstruct_uid).glob("*.dcm"))

    pre = "Test_"

    output_img = "img.nii.gz"
    with tempfile.TemporaryDirectory() as temp_dir:
        output_dir = Path(temp_dir)

        # Run the function
        convert_rtstruct(
            ct_path, rtstruct_path, prefix=pre, output_dir=output_dir, output_img=output_img
        )

        # Check some of the output files for sanity
        assert len(list(output_dir.glob("*.nii.gz"))) == 6

        # Check the converted image series
        ct_output = next(output_dir.glob("img.nii.gz"))
        im = sitk.ReadImage(str(ct_output))

        assert np.allclose(im.GetOrigin(), (-249.51, -483.01, -640.2), atol=0.01)
        assert np.allclose(im.GetSize(), (512, 512, 130), atol=0.01)
        assert np.allclose(im.GetSpacing(), (0.97, 0.97, 3.0), atol=0.01)

        nda = sitk.GetArrayFromImage(im)
        assert nda.sum() == -23952778432

        # Check a converted contour mask
        heart_output = next(output_dir.glob("Test_Heart.nii.gz"))
        mask = sitk.ReadImage(str(heart_output))
        assert np.allclose(mask.GetOrigin(), (-249.51, -483.01, -640.2), atol=0.01)
        assert np.allclose(mask.GetSize(), (512, 512, 130), atol=0.01)
        assert np.allclose(mask.GetSpacing(), (0.97, 0.97, 3.0), atol=0.01)
        nda = sitk.GetArrayFromImage(mask)

        assert nda.sum() == 263237


def test_convert_nifti_to_dicom(nifti_data, dicom_data):

    test_pat_id = "LCTSC-Test-S1-101"
    ct_uid = "1.3.6.1.4.1.14519.5.2.1.7014.4598.106943890850011666503487579262"
    rtstruct_uid = "1.3.6.1.4.1.14519.5.2.1.7014.4598.280355341349691222365783556597"

    pat_path = dicom_data.joinpath(test_pat_id)
    ct_path = pat_path.joinpath(ct_uid)
    rtstruct_path = next(pat_path.joinpath(rtstruct_uid).glob("*.dcm"))

    # Generate dict of masks in the masks directory
    masks = {}
    for mask_path in nifti_data.joinpath(test_pat_id).glob("STRUCTURES/*.nii.gz"):
        name = mask_path.name.split(".")[0].split("RTSTRUCT_")[1]
        masks[name] = str(mask_path)

    with tempfile.TemporaryDirectory() as temp_dir:

        output_file = Path(temp_dir).joinpath("test.dcm")

        # Run the function
        convert_nifti(ct_path, masks, output_file)

        # Check some of the output files for sanity
        original = pydicom.read_file(rtstruct_path)
        rts = pydicom.read_file(output_file)

        assert rts.Modality == original.Modality
        assert rts.PatientName == original.PatientName
        assert rts.PatientID == original.PatientID

        # Check that all the contour names match
        contour_map = {}
        for i in original.StructureSetROISequence:
            for j in rts.StructureSetROISequence:
                if j.ROIName.upper() == i.ROIName.upper():
                    contour_map[int(i.ROINumber)] = int(j.ROINumber)

        assert len(contour_map.keys()) == 5

        # Confirm that the min/max contour points land within 1 voxel size of the original
        i = 1
        for j in original.ROIContourSequence[i - 1].ContourSequence:

            for k in rts.ROIContourSequence[contour_map[i] - 1].ContourSequence:

                if j.ContourData[2] == k.ContourData[2]:
                    j_points = np.array(j.ContourData)
                    j_points = j_points.reshape(j.NumberOfContourPoints, 3)

                    k_points = np.array(k.ContourData)
                    k_points = k_points.reshape(k.NumberOfContourPoints, 3)
                    assert (max(abs(k_points.min(axis=0) - j_points.min(axis=0)))) < 0.83
                    assert (max(abs(k_points.max(axis=0) - j_points.max(axis=0)))) < 0.83



import numpy as np
import SimpleITK as sitk

from scipy.interpolate import griddata

from platipy.imaging.label.utils import vectorised_transform_index_to_physical_point


def evaluate_distance_on_surface(
    reference_volume, test_volume, abs_distance=True, reference_as_distance_map=False
):
    """
    Evaluates a distance map on a surface
    Input: reference_volume: binary volume SimpleITK image, or alternatively a distance map
           test_volume: binary volume SimpleITK image
    Output: theta, phi, values
    """
    if reference_as_distance_map:
        reference_distance_map = reference_volume
    else:
        if abs_distance:
            reference_distance_map = sitk.Abs(
                sitk.SignedMaurerDistanceMap(
                    reference_volume, squaredDistance=False, useImageSpacing=True
                )
            )

        else:
            reference_distance_map = sitk.SignedMaurerDistanceMap(
                reference_volume, squaredDistance=False, useImageSpacing=True
            )

    test_surface = sitk.LabelContour(test_volume)

    distance_image = sitk.Multiply(
        reference_distance_map, sitk.Cast(test_surface, sitk.sitkFloat32)
    )
    distance_array = sitk.GetArrayFromImage(distance_image)

    # Get centre of mass of reference volume
    reference_volume_array = sitk.GetArrayFromImage(reference_volume)
    reference_volume_locations = np.where(reference_volume_array == 1)
    com_index = reference_volume_locations.mean(axis=1)
    com_real = vectorised_transform_index_to_physical_point(reference_volume, com_index)

    # Calculate centre of mass in real coordinates
    test_surface_array = sitk.GetArrayFromImage(test_surface)
    test_surface_locations = np.where(test_surface_array == 1)
    test_surface_locations_array = np.array(test_surface_locations)

    # Calculate each point on the surface in real coordinates
    pts = test_surface_locations_array.T
    pts_real = vectorised_transform_index_to_physical_point(test_surface, pts)
    pts_diff = pts_real - com_real

    # Convert to spherical polar coordinates - base at north pole
    rho = np.sqrt((pts_diff * pts_diff).sum(axis=1))
    theta = np.pi / 2.0 - np.arccos(pts_diff.T[0] / rho)
    phi = -1 * np.arctan2(pts_diff.T[2], -1.0 * pts_diff.T[1])

    # Extract values
    values = distance_array[test_surface_locations]

    return theta, phi, values


def evaluate_distance_to_reference(reference_volume, test_volume, resample_factor=1):
    """
    Evaluates the distance from the surface of a test volume to a reference
    Input: reference_volume: binary volume SimpleITK image
           test_volume: binary volume SimpleITK image
    Output: values : the distance to each point on the reference volume surface
    """

    # TO DO
    # come up with a better resampling strategy
    # e.g. resample images prior to this process?

    # compute the distance map from the test volume surface
    test_distance_map = sitk.Abs(
        sitk.SignedMaurerDistanceMap(test_volume, squaredDistance=False, useImageSpacing=True)
    )

    # get the distance from the test surface to the reference surface
    ref_surface = sitk.LabelContour(reference_volume)
    ref_surface_pts = sitk.GetArrayFromImage(ref_surface) == 1
    surface_values = sitk.GetArrayFromImage(test_distance_map)[ref_surface_pts]

    # resample to keep the points to a reasonable amount
    values = surface_values[::resample_factor]

    return values


def regrid_spherical_data(theta, phi, values, resolution):
    """
    Re-grids spherical data
    Input: theta, phi, values
    Options: plot a figure (plotFig), save a figure (saveFig), case identifier (figName)
    Output: p_lat, p_long, grid_values (, fig)
    """
    # Re-grid:
    #  Set up grid
    d_radian = resolution * np.pi / 180
    p_long, p_lat = np.mgrid[-np.pi : np.pi : d_radian, -np.pi / 2.0 : np.pi / 2.0 : d_radian]

    # First pass - linear interpolation, works well but not for edges
    grid_values = griddata(
        list(zip(theta, phi)), values, (p_lat, p_long), method="linear", rescale=False
    )

    # Second pass - nearest neighbour interpolation
    grid_values_nn = griddata(
        list(zip(theta, phi)), values, (p_lat, p_long), method="nearest", rescale=False
    )

    # Third pass - wherever the linear interpolation isn't defined use nearest neighbour
    # interpolation
    grid_values[~np.isfinite(grid_values)] = grid_values_nn[~np.isfinite(grid_values)]

    return p_lat, p_long, grid_values



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import SimpleITK as sitk

from platipy.imaging.registration.utils import (
    apply_transform,
    convert_mask_to_reg_structure,
)

from platipy.imaging.registration.deformable import (
    fast_symmetric_forces_demons_registration,
)


def generate_field_shift(mask_image, vector_shift=(10, 10, 10), gaussian_smooth=5):
    """
    Shifts (moves) a structure defined using a binary mask.

    Args:
        mask_image ([SimpleITK.Image]): The binary mask to shift.
        vector_shift (tuple, optional): The displacement vector applied to the entire binary mask.
                                        Convention: (+/-, +/-, +/-) = (sup/inf, post/ant,
                                        left/right) shift.
                                        Defined in millimetres.
                                        Defaults to (10, 10, 10).
        gaussian_smooth (int | list, optional): Scale of a Gaussian kernel used to smooth the
                                                deformation vector field. Defaults to 5.

    Returns:
        [SimpleITK.Image]: The binary mask following the shift.
        [SimpleITK.DisplacementFieldTransform]: The transform representing the shift.
        [SimpleITK.Image]: The displacement vector field representing the shift.
    """
    # Define array
    # Used for image array manipulations
    mask_image_arr = sitk.GetArrayFromImage(mask_image)

    # The template deformation field
    # Used to generate transforms
    dvf_arr = np.zeros(mask_image_arr.shape + (3,))
    dvf_arr = dvf_arr - np.array([[[vector_shift[::-1]]]])
    dvf_template = sitk.GetImageFromArray(dvf_arr)

    # Copy image information
    dvf_template.CopyInformation(mask_image)

    dvf_tfm = sitk.DisplacementFieldTransform(sitk.Cast(dvf_template, sitk.sitkVectorFloat64))
    mask_image_shift = apply_transform(
        mask_image, transform=dvf_tfm, default_value=0, interpolator=sitk.sitkNearestNeighbor
    )

    dvf_template = sitk.Mask(dvf_template, mask_image | mask_image_shift)

    # smooth
    if np.any(gaussian_smooth):

        if not hasattr(gaussian_smooth, "__iter__"):
            gaussian_smooth = (gaussian_smooth,) * 3

        dvf_template = sitk.SmoothingRecursiveGaussian(dvf_template, gaussian_smooth)

    dvf_tfm = sitk.DisplacementFieldTransform(sitk.Cast(dvf_template, sitk.sitkVectorFloat64))
    mask_image_shift = apply_transform(
        mask_image, transform=dvf_tfm, default_value=0, interpolator=sitk.sitkNearestNeighbor
    )

    return mask_image_shift, dvf_tfm, dvf_template


def generate_field_asymmetric_contract(
    mask_image, vector_asymmetric_contract=(10, 10, 10), gaussian_smooth=5, compute_real_dvf=False
):
    """
    Contracts a structure (defined using a binary mask) using a specified vector.

    Args:
        mask_image ([SimpleITK.Image]): The binary mask to contract.
        vector_asymmetric_contract (tuple, optional): The contraction vector applied to the entire
                                                      binary mask.
                                                      Convention: (+/-, +/-, +/-) = (sup/inf,
                                                      post/ant, left/right) border is contracted.
                                                      Defined in millimetres.
                                                      Defaults to (10, 10, 10).
        gaussian_smooth (int | list, optional): Scale of a Gaussian kernel used to smooth the
                                                deformation vector field. Defaults to 5.
        compute_real_dvf (bool, optional): If True, the real deformation vector field is computed.
                                           This involves a slower computation. Defaults to False.

    Returns:
        [SimpleITK.Image]: The binary mask following the contract.
        [SimpleITK.DisplacementFieldTransform]: The transform representing the contract.
        [SimpleITK.Image]: The displacement vector field representing the contract.
    """
    # Define array
    # Used for image array manipulations
    mask_image_arr = sitk.GetArrayFromImage(mask_image)

    # The template deformation field
    # Used to generate transforms
    dvf_arr = np.zeros(mask_image_arr.shape + (3,))
    dvf_arr = dvf_arr + np.array([[[vector_asymmetric_contract[::-1]]]])
    dvf_template = sitk.GetImageFromArray(dvf_arr)

    # Copy image information
    dvf_template.CopyInformation(mask_image)

    dvf_template = sitk.Mask(dvf_template, mask_image)

    dvf_tfm = sitk.DisplacementFieldTransform(sitk.Cast(dvf_template, sitk.sitkVectorFloat64))

    mask_image_asymmetric_contract = apply_transform(
        mask_image, transform=dvf_tfm, default_value=0, interpolator=sitk.sitkNearestNeighbor
    )

    if compute_real_dvf:

        reg_struct = convert_mask_to_reg_structure(mask_image, expansion=3)
        reg_struct_def = convert_mask_to_reg_structure(mask_image_asymmetric_contract, expansion=3)

        # Use DSGR to compute the realistic deformation vector field
        _, _, dvf_template = fast_symmetric_forces_demons_registration(
            reg_struct_def,
            reg_struct,
            isotropic_resample=True,
            resolution_staging=[4, 2],
            iteration_staging=[20, 10],
        )

    # smooth
    if np.any(gaussian_smooth):

        if not hasattr(gaussian_smooth, "__iter__"):
            gaussian_smooth = (gaussian_smooth,) * 3

        dvf_template = sitk.SmoothingRecursiveGaussian(dvf_template, gaussian_smooth)

    dvf_tfm = sitk.DisplacementFieldTransform(sitk.Cast(dvf_template, sitk.sitkVectorFloat64))
    mask_image_asymmetric_contract = apply_transform(
        mask_image, transform=dvf_tfm, default_value=0, interpolator=sitk.sitkNearestNeighbor
    )

    return mask_image_asymmetric_contract, dvf_tfm, dvf_template


def generate_field_asymmetric_extend(
    mask_image, vector_asymmetric_extend=(10, 10, 10), gaussian_smooth=5
):
    """
    Extends a structure (defined using a binary mask) using a specified vector.

    Args:
        mask_image ([SimpleITK.Image]): The binary mask to extend.
        vector_asymmetric_extend (tuple, optional): The extension vector applied to the entire
                                                    binary mask.
                                                    Convention: (+/-, +/-, +/-) = (sup/inf,
                                                    post/ant, left/right) border is extended.
                                                    Defined in millimetres.
                                                    Defaults to (10, 10, 10).
        gaussian_smooth (int | list, optional): Scale of a Gaussian kernel used to smooth the
                                                deformation vector field. Defaults to 5.

    Returns:
        [SimpleITK.Image]: The binary mask following the extension.
        [SimpleITK.DisplacementFieldTransform]: The transform representing the extension.
        [SimpleITK.Image]: The displacement vector field representing the extension.
    """
    # Define array
    # Used for image array manipulations
    mask_image_arr = sitk.GetArrayFromImage(mask_image)

    # The template deformation field
    # Used to generate transforms
    dvf_arr = np.zeros(mask_image_arr.shape + (3,))
    dvf_arr = dvf_arr - np.array([[[vector_asymmetric_extend[::-1]]]])
    dvf_template = sitk.GetImageFromArray(dvf_arr)

    # Copy image information
    dvf_template.CopyInformation(mask_image)

    dvf_tfm = sitk.DisplacementFieldTransform(sitk.Cast(dvf_template, sitk.sitkVectorFloat64))

    mask_image_asymmetric_extend = apply_transform(
        mask_image, transform=dvf_tfm, default_value=0, interpolator=sitk.sitkNearestNeighbor
    )

    dvf_template = sitk.Mask(dvf_template, mask_image_asymmetric_extend)

    # smooth
    if np.any(gaussian_smooth):

        if not hasattr(gaussian_smooth, "__iter__"):
            gaussian_smooth = (gaussian_smooth,) * 3

        dvf_template = sitk.SmoothingRecursiveGaussian(dvf_template, gaussian_smooth)

    dvf_tfm = sitk.DisplacementFieldTransform(sitk.Cast(dvf_template, sitk.sitkVectorFloat64))

    mask_image_asymmetric_extend = apply_transform(
        mask_image, transform=dvf_tfm, default_value=0, interpolator=sitk.sitkNearestNeighbor
    )

    return mask_image_asymmetric_extend, dvf_tfm, dvf_template


def generate_field_expand(
    mask,
    bone_mask=False,
    expand=3,
    gaussian_smooth=5,
    use_internal_deformation=True,
):
    """
    Expands a structure (defined using a binary mask) using a specified vector to define the
    dilation kernel.

    Args:
        mask ([SimpleITK.Image]): The binary mask to expand.
        bone_mask ([SimpleITK.Image, optional]): A binary mask defining regions where we expect
                                                 restricted deformations.
        vector_asymmetric_extend (int |tuple, optional): The expansion vector applied to the entire
                                                         binary mask.
                                                    Convention: (z,y,x) size of expansion kernel.
                                                    Defined in millimetres.
                                                    Defaults to 3.
        gaussian_smooth (int | list, optional): Scale of a Gaussian kernel used to smooth the
                                                deformation vector field. Defaults to 5.

    Returns:
        [SimpleITK.Image]: The binary mask following the expansion.
        [SimpleITK.DisplacementFieldTransform]: The transform representing the expansion.
        [SimpleITK.Image]: The displacement vector field representing the expansion.
    """

    if bone_mask is not False:
        mask_original = mask + bone_mask
    else:
        mask_original = mask

    # Use binary erosion to create a smaller volume
    if not hasattr(expand, "__iter__"):
        expand = (expand,) * 3

    expand = np.array(expand)

    # Convert voxels to millimetres
    expand = expand / np.array(mask.GetSpacing()[::-1])

    # Re-order to (x,y,z)
    expand = expand[::-1]
    # expand = [int(i / j) for i, j in zip(expand, mask.GetSpacing()[::-1])][::-1]

    # If all negative: erode
    if np.all(np.array(expand) <= 0):
        print("All factors negative: shrinking only.")
        mask_expand = sitk.BinaryErode(mask, np.abs(expand).astype(int).tolist(), sitk.sitkBall)

    # If all positive: dilate
    elif np.all(np.array(expand) >= 0):
        print("All factors positive: expansion only.")
        mask_expand = sitk.BinaryDilate(mask, np.abs(expand).astype(int).tolist(), sitk.sitkBall)

    # Otherwise: sequential operations
    else:
        print("Mixed factors: shrinking and expansion.")
        expansion_kernel = expand * (expand > 0)
        shrink_kernel = expand * (expand < 0)

        mask_expand = sitk.BinaryDilate(
            mask, np.abs(expansion_kernel).astype(int).tolist(), sitk.sitkBall
        )
        mask_expand = sitk.BinaryErode(
            mask_expand, np.abs(shrink_kernel).astype(int).tolist(), sitk.sitkBall
        )

    if bone_mask is not False:
        mask_expand = mask_expand + bone_mask

    if use_internal_deformation:
        registration_mask_original = convert_mask_to_reg_structure(mask_original)
        registration_mask_expand = convert_mask_to_reg_structure(mask_expand)

    else:
        registration_mask_original = mask_original
        registration_mask_expand = mask_expand

    # Use DIR to find the deformation
    _, _, dvf_template = fast_symmetric_forces_demons_registration(
        registration_mask_expand,
        registration_mask_original,
        isotropic_resample=True,
        resolution_staging=[4, 2],
        iteration_staging=[10, 10],
        ncores=8,
    )

    # smooth
    if np.any(gaussian_smooth):

        if not hasattr(gaussian_smooth, "__iter__"):
            gaussian_smooth = (gaussian_smooth,) * 3

        dvf_template = sitk.SmoothingRecursiveGaussian(dvf_template, gaussian_smooth)

    dvf_tfm = sitk.DisplacementFieldTransform(sitk.Cast(dvf_template, sitk.sitkVectorFloat64))

    mask_symmetric_expand = apply_transform(
        mask, transform=dvf_tfm, default_value=0, interpolator=sitk.sitkNearestNeighbor
    )

    return mask_symmetric_expand, dvf_tfm, dvf_template


def generate_field_radial_bend(
    reference_image,
    body_mask,
    reference_point,
    axis_of_rotation=[0, 0, -1],
    scale=0.1,
    mask_bend_from_reference_point=("z", "inf"),
    gaussian_smooth=5,
):
    """
    Generates a synthetic field characterised by radial bending.
    Typically, this field would be used to simulate a moving head and so masking is important.

    Args:
        reference_image ([SimpleITK.Image]): The image to be deformed.
        body_mask ([SimpleITK.Image]): A binary mask in which the deformation field will be defined
        reference_point ([tuple]): The point (z,y,x) about which the rotation field is defined.
        axis_of_rotation (tuple, optional): The axis of rotation (z,y,x). Defaults to [0, 0, -1].
        scale (int, optional): The deformation vector length at each point will equal scale
                              multiplied by the distance to that point from reference_point.
                              Defaults to 1.
        mask_bend_from_reference_point (tuple, optional): The dimension (z=axial, y=coronal,
                                                          x=sagittal) and limit (inf/sup, post/ant,
                                                          left/right) for masking the vector field,
                                                          relative to the reference point. Defaults
                                                          to ("z", "inf").
        gaussian_smooth (int | list, optional): Scale of a Gaussian kernel used to smooth the
                                                deformation vector field. Defaults to 5.

    Returns:
        [SimpleITK.Image]: The binary mask following the expansion.
        [SimpleITK.DisplacementFieldTransform]: The transform representing the expansion.
        [SimpleITK.Image]: The displacement vector field representing the expansion.
    """

    body_mask_arr = sitk.GetArrayFromImage(body_mask)

    if mask_bend_from_reference_point is not False:
        if mask_bend_from_reference_point[0] == "z":
            if mask_bend_from_reference_point[1] == "inf":
                body_mask_arr[: reference_point[0], :, :] = 0
            elif mask_bend_from_reference_point[1] == "sup":
                body_mask_arr[reference_point[0] :, :, :] = 0
        if mask_bend_from_reference_point[0] == "y":
            if mask_bend_from_reference_point[1] == "post":
                body_mask_arr[:, reference_point[1] :, :] = 0
            elif mask_bend_from_reference_point[1] == "ant":
                body_mask_arr[:, : reference_point[1], :] = 0
        if mask_bend_from_reference_point[0] == "x":
            if mask_bend_from_reference_point[1] == "left":
                body_mask_arr[:, :, reference_point[2] :] = 0
            elif mask_bend_from_reference_point[1] == "right":
                body_mask_arr[:, :, : reference_point[2]] = 0

    pt_arr = np.array(np.where(body_mask_arr))
    vector_ref_to_pt = pt_arr - np.array(reference_point)[:, None]

    # Normalise the normal vector (axis_of_rotation)
    axis_of_rotation = np.array(axis_of_rotation)
    axis_of_rotation = axis_of_rotation / np.linalg.norm(axis_of_rotation)

    deformation_vectors = np.cross(vector_ref_to_pt[::-1].T, axis_of_rotation[::-1])

    dvf_template = sitk.Image(reference_image.GetSize(), sitk.sitkVectorFloat64, 3)
    dvf_template_arr = sitk.GetArrayFromImage(dvf_template)

    if scale is not False:
        dvf_template_arr[np.where(body_mask_arr)] = deformation_vectors * scale

    dvf_template = sitk.GetImageFromArray(dvf_template_arr)
    dvf_template.CopyInformation(reference_image)

    # smooth
    if np.any(gaussian_smooth):

        if not hasattr(gaussian_smooth, "__iter__"):
            gaussian_smooth = (gaussian_smooth,) * 3

        dvf_template = sitk.SmoothingRecursiveGaussian(dvf_template, gaussian_smooth)

    dvf_tfm = sitk.DisplacementFieldTransform(sitk.Cast(dvf_template, sitk.sitkVectorFloat64))
    reference_image_bend = apply_transform(
        reference_image,
        transform=dvf_tfm,
        default_value=int(sitk.GetArrayViewFromImage(reference_image).min()),
        interpolator=sitk.sitkLinear,
    )

    return reference_image_bend, dvf_tfm, dvf_template



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import logging

import numpy as np
import SimpleITK as sitk

from scipy.optimize import curve_fit
from scipy.ndimage import filters
from scipy.stats import norm as scipy_norm

from platipy.imaging.label.fusion import combine_labels, process_probability_image

from platipy.imaging.label.projection import (
    evaluate_distance_on_surface,
    evaluate_distance_to_reference,
    regrid_spherical_data,
)

logger = logging.getLogger(__name__)

def median_absolute_deviation(data, axis=None):
    """Median Absolute Deviation: a "Robust" version of standard deviation.
    Indices variabililty of the sample.
    https://en.wikipedia.org/wiki/Median_absolute_deviation
    """
    return np.median(np.abs(data - np.median(data, axis=axis)), axis=axis)


def gaussian_curve(x, a, m, s):
    """Returns a Gaussian (normal) curve

    Args:
        x (np.ndarray): values to sample the normal curve
        a (float): magnitude
        m (float): location (mean)
        s (float): scale (standard deviation)

    Returns:
        np.ndarray: sampled values along the normal curve
    """
    return a * scipy_norm.pdf(x, loc=m, scale=s)


def run_iar(
    atlas_set,
    reference_structure,
    smooth_distance_maps=False,
    smooth_sigma=1,
    z_score_statistic="MAD",
    outlier_method="IQR",
    min_best_atlases=10,
    outlier_factor=1.5,
    iteration=0,
    single_step=False,
    project_on_sphere=False,
    label="DIR",
):
    """
    Perform iterative atlas removal on the atlas_set
    """

    if iteration == 0:
        # Run some checks in the data
        # If there is a MAJOR error we need to check

        # Begin the process
        logger.info("Iterative atlas removal: ")
        logger.info("  Beginning process")

    # Get remaining case identifiers to loop through
    remaining_id_list = list(atlas_set.keys())

    # Generate the surface projections
    #   1. Set the consensus surface using the reference volume
    probability_label = combine_labels(atlas_set, reference_structure, label=label)[
        reference_structure
    ]

    # Modify resolution for better statistics
    if project_on_sphere:
        if len(remaining_id_list) < 12:
            logger.info("  Less than 12 atlases, resolution set: 3x3 sqr deg")
            resolution = 3
        elif len(remaining_id_list) < 7:
            logger.info("  Less than 7 atlases, resolution set: 6x6 sqr deg")
            resolution = 6
        else:
            resolution = 1
    else:
        if len(remaining_id_list) < 12:
            logger.info("  Less than 12 atlases, resample factor set: 5")
            resample_factor = 5
        elif len(remaining_id_list) < 7:
            logger.info("  Less than 7 atlases, resolution set: 6x6 sqr deg")
            resample_factor = 10
        else:
            resample_factor = 1

    g_val_list = []
    logger.info("  Calculating surface distance maps: ")
    for test_id in remaining_id_list:
        logger.info("    %s", test_id)
        #   2. Calculate the distance from the surface to the consensus surface

        test_volume = atlas_set[test_id][label][reference_structure]

        # This next step ensures non-binary labels are treated properly
        # We use 0.1 to capture the outer edge of the test delineation, if it is probabilistic
        test_volume = process_probability_image(test_volume, 0.1)

        if project_on_sphere:
            reference_volume = process_probability_image(probability_label, threshold=0.999)
            # note: we use a threshold slightly below 1 to ensure the consensus (reference) volume
            # is a suitable binary volume

            # Compute the reference distance map
            reference_distance_map = sitk.Abs(
                sitk.SignedMaurerDistanceMap(
                    reference_volume, squaredDistance=False, useImageSpacing=True
                )
            )

            # Compute the distance to test surfaces, across the surface of the reference
            theta, phi, values = evaluate_distance_on_surface(
                reference_distance_map, test_volume, reference_as_distance_map=True
            )

            _, _, g_vals = regrid_spherical_data(theta, phi, values, resolution=resolution)

            g_val_list.append(g_vals)
        else:
            reference_volume = process_probability_image(probability_label, threshold=0.95)
            # note: we use a threshold slightly below 1 to ensure the consensus (reference) volume
            # is a suitable binary volume we have the flexibility to modify the reference volume
            # when we do not use spherical projection a larger surface means more evaluations and
            # better statistics, so we prefer a lower threshold but not too low,
            # or it may include some errors

            # Compute distance to reference, from the test volume
            values = evaluate_distance_to_reference(
                reference_volume, test_volume, resample_factor=resample_factor
            )

            g_val_list.append(values)

    q_results = {}

    for i, (test_id, g_vals) in enumerate(zip(remaining_id_list, g_val_list)):

        g_val_list_test = g_val_list[:]
        g_val_list_test.pop(i)

        if project_on_sphere and smooth_distance_maps:
            g_vals = filters.gaussian_filter(g_vals, sigma=smooth_sigma, mode="wrap")

        #       b) i] Compute the Z-scores over the projected surface
        if z_score_statistic.lower() == "std":
            g_val_mean = np.mean(g_val_list_test, axis=0)
            g_val_std = np.std(g_val_list_test, axis=0)

            if np.any(g_val_std == 0):
                logger.info("    Std Dev zero count: %d", np.sum(g_val_std == 0))
                g_val_std[g_val_std == 0] = g_val_std.mean()

            z_score_vals_array = (g_vals - g_val_mean) / g_val_std

        elif z_score_statistic.lower() == "mad":
            g_val_median = np.median(g_val_list_test, axis=0)
            g_val_mad = 1.4826 * median_absolute_deviation(g_val_list_test, axis=0)

            if np.any(~np.isfinite(g_val_mad)):
                logger.info("Error in MAD")
                logger.info(g_val_mad)

            if np.any(g_val_mad == 0):
                logger.info("    MAD zero count: %d", np.sum(g_val_mad == 0))
                g_val_mad[g_val_mad == 0] = np.median(g_val_mad)

            z_score_vals_array = (g_vals - g_val_median) / g_val_mad

        else:
            logger.error(" Error!")
            logger.error(" z_score must be one of: MAD, STD")
            raise ValueError("z_score must be one of: MAD, STD")

        z_score_vals = np.ravel(z_score_vals_array)

        logger.debug("      [%s] Statistics of mZ-scores", test_id)
        logger.debug("        Min(Z)    = %.2f", z_score_vals.min())
        logger.debug("        Q1(Z)     = %.2f", np.percentile(z_score_vals, 25))
        logger.debug("        Mean(Z)   = %.2f", z_score_vals.mean())
        logger.debug("        Median(Z) = %.2f", np.percentile(z_score_vals, 50))
        logger.debug("        Q3(Z)     = %.2f", np.percentile(z_score_vals, 75))
        logger.debug("        Max(Z)    = %.2f\n", z_score_vals.max())

        # Calculate excess area from Gaussian: the Q-metric
        bins = np.linspace(-15, 15, 501)
        z_density, bin_edges = np.histogram(z_score_vals, bins=bins, density=True)
        bin_centers = (bin_edges[1:] + bin_edges[:-1]) / 2.0

        try:
            popt, _ = curve_fit(  # pylint: disable=unbalanced-tuple-unpacking
                f=gaussian_curve, xdata=bin_centers, ydata=z_density
            )

            z_ideal = gaussian_curve(bin_centers, *popt)
            z_diff = np.abs(z_density - z_ideal)
        except (RuntimeError, ValueError):
            logger.debug("IAR couldnt fit curve, estimating with sampled statistics.")
            z_ideal = gaussian_curve(bin_centers, a=1, m=z_density.mean(), s=z_density.std())
            z_diff = np.abs(z_density - z_ideal)

        # Integrate to get the q_value
        q_value = np.trapz(z_diff * np.abs(bin_centers) ** 2, bin_centers)
        q_results[test_id] = np.float64(q_value)

    # Exclude (at most) the worst 3 atlases for outlier detection
    # With a minimum number, this helps provide more robust estimates at low numbers
    result_list = list(q_results.values())
    result_list = [r for r in result_list if ~np.isnan(r) and np.isfinite(r)]
    best_results = np.sort(result_list)[: max([min_best_atlases, len(result_list) - 3])]

    if outlier_method.lower() == "iqr":
        outlier_limit = np.percentile(best_results, 75, axis=0) + outlier_factor * np.subtract(
            *np.percentile(best_results, [75, 25], axis=0)
        )
    elif outlier_method.lower() == "std":
        outlier_limit = np.mean(best_results, axis=0) + outlier_factor * np.std(
            best_results, axis=0
        )
    else:
        logger.error(" Error!")
        logger.error(" outlier_method must be one of: IQR, STD")
        sys.exit()

    logger.info("  Analysing results")
    logger.info("   Outlier limit: %6.3f}", outlier_limit)
    keep_id_list = []

    logger.info(
        "%s,%s,%s,%.4g}\n",
            iteration,
            " ".join(remaining_id_list),
            " ".join(["{0:.4g}".format(i) for i in list(q_results.values())]),
            outlier_limit,
    )

    for idx, result in q_results.items():

        accept = result <= outlier_limit

        logger.info(
            "      %s: Q = %6.3f [%s]",
                idx, result, {True: "KEEP", False: "REMOVE"}[accept]
        )

        if accept:
            keep_id_list.append(idx)

    if len(keep_id_list) < len(remaining_id_list):
        logger.info("\n  Step %d Complete", iteration)
        logger.info("  Num. Removed = %d --\n", len(remaining_id_list) - len(keep_id_list))

        iteration += 1
        atlas_set_new = {i: atlas_set[i] for i in keep_id_list}

        if single_step:
            return atlas_set_new

        return run_iar(
            atlas_set=atlas_set_new,
            reference_structure=reference_structure,
            smooth_distance_maps=smooth_distance_maps,
            smooth_sigma=smooth_sigma,
            z_score_statistic=z_score_statistic,
            outlier_method=outlier_method,
            min_best_atlases=min_best_atlases,
            outlier_factor=outlier_factor,
            iteration=iteration,
            project_on_sphere=project_on_sphere,
            label=label,
        )

    logger.info("  End point reached. Keeping:\n   %s", keep_id_list)

    return atlas_set



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import zipfile
import os

from pathlib import Path

import requests
import logging

logger = logging.getLogger(__name__)

from platipy.dicom.io.crawl import (
    process_dicom_directory,
)

API_URL = "https://services.cancerimagingarchive.net/services/v4/TCIA"

collection_endpoint = f"{API_URL}/query/getCollectionValues"
modalities_endpoint = f"{API_URL}/query/getModalityValues"
patient_endpoint = f"{API_URL}/query/getPatient"
series_endpoint = f"{API_URL}/query/getSeries"
series_size_endpoint = f"{API_URL}/query/getSeriesSize"
download_series_endpoint = f"{API_URL}/query/getImage"
sop_uids_endpoint = f"{API_URL}/query/getSOPInstanceUIDs"
download_image_endpoint = f"{API_URL}/query/getSingleImage"


def get_collections():
    """Get a list of collections available in the TCIA database

    Returns:
        list: List of collections
    """

    res = requests.get(collection_endpoint)
    collections = [collection["Collection"] for collection in res.json()]
    collections.sort()

    return collections


def get_patients_in_collection(collection):
    """Get a list of sorted patient IDs in a TCIA collection

    Args:
        collection (str): TCIA collection

    Returns:
        list: List of sorted Patient IDs in the collection
    """

    res = requests.get(patient_endpoint, params={"Collection": collection})
    patient_ids = [pat["PatientID"] for pat in res.json()]
    patient_ids.sort()

    return patient_ids


def get_modalities_in_collection(collection):
    """Get a list of object modalities in a TCIA collection

    Args:
        collection (str): TCIA collection

    Returns:
        list: List of modalities available in the collection
    """

    res = requests.get(modalities_endpoint, params={"Collection": collection})
    modalities = [mod["Modality"] for mod in res.json()]

    return modalities


def get_lung_data(number_of_patients=1):
    """Gets some images and structures from the LCTSC dataset in TCIA.

    Keyword Arguments:
        number_of_patients {int} -- The number of patients to fetch (default: {1})

    Returns:
        dict -- dictionary of patient ids as key and path to data as value
    """

    collection = "LCTSC"
    patient_ids = get_patients_in_collection(collection)
    return fetch_data(
        collection, patient_ids=patient_ids[0:number_of_patients], modalities=["CT", "RTSTRUCT"]
    )


def get_hn_data(number_of_patients=1):
    """Gets some images and structures from the Head-Neck-Radiomics-HN1 dataset in TCIA.

    Keyword Arguments:
        number_of_patients {int} -- The number of patients to fetch (default: {1})

    Returns:
        dict -- dictionary of patient ids as key and path to data as value
    """

    collection = "HEAD-NECK-RADIOMICS-HN1"
    patient_ids = get_patients_in_collection(collection)
    return fetch_data(
        collection, patient_ids=patient_ids[0:number_of_patients], modalities=["CT", "RTSTRUCT"]
    )


def fetch_data(
    collection, patient_ids=None, modalities=None, nifti=True, output_directory="./tcia"
):
    """Fetches data from TCIA from the dataset specified

    Args:
        collection (str): The TCIA collection to fetch from
        patient_ids (list, optional): The patient IDs to fetch. If not set all patients are
                                      fetched
        modalities (list, optional): A list of strings definiing the modalites to fetch. Will fetch
                                     all modalities available if not specified.
        nifti (bool, optional): Whether or not to convert the fetched DICOM to Nifti. Defaults to
                                True
        output_directory (str): The directory in which to place fetched data

    Returns:
        dict: The patients and directories where their data was fetched
    """

    result = {}

    if isinstance(output_directory, str):
        output_directory = Path(output_directory)

    output_directory = output_directory.joinpath(collection)
    output_directory.mkdir(exist_ok=True, parents=True)

    modalities_available = get_modalities_in_collection(collection)
    logger.debug("Modalities available: %s", modalities_available)

    if modalities is None:
        logger.debug("Will fetch all modalities in collection")
        modalities = modalities_available
    else:
        # Check that the user supplied modalities are all available, raise error if not
        modalities_all_available = True
        for modality in modalities:
            if not modality in modalities_available:
                modalities_all_available = False
                logger.error("Modality not available in collection: %s", modality)

        if not modalities_all_available:
            raise ValueError("Modalities aren't all available in collection")

    if not patient_ids:
        patient_ids = get_patients_in_collection(collection)

    for pid in patient_ids:

        patient_directory = output_directory.joinpath(pid)
        dicom_directory = patient_directory.joinpath("DICOM")
        nifti_directory = patient_directory.joinpath("NIFTI")
        result[pid] = {}
        result[pid]["DICOM"] = {}

        logger.debug("Fetching data for Patient: %s", pid)

        for modality in modalities:
            res = requests.get(
                series_endpoint,
                params={"Collection": collection, "PatientID": pid, "Modality": modality},
            )
            series = json.loads(res.text)

            if not modality in result[pid]:
                result[pid]["DICOM"][modality] = {}

            for obj in series:

                series_uid = obj["SeriesInstanceUID"]

                target_directory = dicom_directory.joinpath(series_uid)
                result[pid]["DICOM"][modality][series_uid] = target_directory
                if target_directory.exists():
                    logger.warning(
                        "Series directory exists: %s, won't fetch data", target_directory
                    )
                    continue

                logger.debug("Downloading Series: %s", series_uid)

                target_directory.mkdir(parents=True)

                save_path = target_directory.joinpath(f"{pid}.zip")

                response = requests.get(
                    download_series_endpoint,
                    stream=True,
                    params={"SeriesInstanceUID": obj["SeriesInstanceUID"]},
                )
                with open(save_path, "wb") as file_obj:
                    for chunk in response.iter_content():
                        file_obj.write(chunk)

                with zipfile.ZipFile(save_path, "r") as zip_ref:
                    zip_ref.extractall(target_directory)

                os.remove(save_path)

        if nifti:
            logger.info("Converting data for %s to Nifti", pid)
            nifti_results = process_dicom_directory(
                dicom_directory, output_directory=nifti_directory
            )
            result[pid]["NIFTI"] = nifti_results[pid]

    return result



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import SimpleITK as sitk
import pandas as pd


def calculate_dvh(dose_grid, label, bins=1001):
    """Calculates a dose-volume histogram

    Args:
        dose_grid (SimpleITK.Image): The dose grid.
        label (SimpleITK.Image): The (binary) label defining a structure.
        bins (int | list | np.ndarray, optional): Passed to np.histogram,
            can be an int (number of bins), or a list (specifying bin edges). Defaults to 1001.

    Returns:
        bins (numpy.ndarray): The points of the dose bins
        values (numpy.ndarray): The DVH values
    """

    if dose_grid.GetSize() != label.GetSize():
        print("Dose grid size does not match label, automatically resampling.")
        dose_grid = sitk.Resample(dose_grid, label)

    dose_arr = sitk.GetArrayViewFromImage(dose_grid)
    label_arr = sitk.GetArrayViewFromImage(label)

    dose_vals = dose_arr[np.where(label_arr)]

    counts, bin_edges = np.histogram(dose_vals, bins=bins)

    # Get mid-points of bins
    bins = (bin_edges[1:] + bin_edges[:-1]) / 2.0

    # Calculate the actual DVH values
    values = np.cumsum(counts[::-1])[::-1]
    if np.all(values == 0):
        return bins, values

    values = values / values.max()

    return bins, values


def calculate_dvh_for_labels(dose_grid, labels, bin_width=0.1, max_dose=None):
    """Calculate the DVH for multiple labels

    Args:
        dose_grid (SimpleITK.Image): Dose grid
        labels (dict): Dictionary of labels with the label name as key and SimpleITK.Image mask as
            value.
        bin_width (float, optional): The width of each bin of the DVH (Gy). Defaults to 0.1.
        max_dose (float, optional): The maximum dose of the DVH. If not set then maximum dose from
            dose grid is used.Defaults to None.

    Returns:
        pandas.DataFrame: The DVH for each structure along with the mean dose and size in cubic
            centimetres as a data frame.
    """

    dvh = []

    label_keys = labels.keys()

    dose_grid = sitk.Resample(dose_grid, labels[list(label_keys)[0]])
    dose_array = sitk.GetArrayFromImage(dose_grid)

    if not max_dose:
        max_dose = dose_array.max()

    for k in label_keys:

        mask = labels[k]
        mask_array = sitk.GetArrayFromImage(mask)

        # Compute cubic centimetre volume of structure
        cc = mask_array.sum() * np.product([a / 10 for a in mask.GetSpacing()])

        bins, values = calculate_dvh(
            dose_grid, labels[k], bins=np.arange(-bin_width / 2, max_dose + bin_width, bin_width)
        )

        # Remove rounding error
        bins = np.round(
            bins.astype(float),
            decimals=10,
        )

        mean_dose = dose_array[mask_array > 0].mean()
        entry = {
            **{
                "label": k,
                "cc": cc,
                "mean": mean_dose,
            },
            **dict(zip(bins, values)),
        }

        dvh.append(entry)

    return pd.DataFrame(dvh)


def calculate_d_x(dvh, x, label=None):
    """Calculate the dose which x percent of the volume receives

    Args:
        dvh (pandas.DataFrame): DVH DataFrame as produced by calculate_dvh_for_labels
        x (float|list): The dose threshold (or list of dose thresholds) which x percent of the
            volume receives
        label (str, optional): The label to compute the metric for. Computes for all metrics if not
            set. Defaults to None.

    Returns:
        pandas.DataFrame: Data frame with a row for each label containing the metric and value.
    """

    if label:
        dvh = dvh[dvh.label == label]

    if not isinstance(x, list):
        x = [x]

    bins = np.array([b for b in dvh.columns if isinstance(b, float)])
    values = np.array(dvh[bins])

    metrics = []
    for idx in range(len(dvh)):
        d = dvh.iloc[idx]

        m = {"label": d.label}

        for threshold in x:
            value = np.interp(threshold / 100, values[idx][::-1], bins[::-1])
            if values[idx, 0] == np.sum(values[idx]):
                value = 0

            # Interp will return zero when computing D100, do compute this separately
            if threshold == 100:
                i, j = np.where(values == 1.0)
                value = bins[j][i == idx][-1]

            m[f"D{threshold}"] = value

        metrics.append(m)

    return pd.DataFrame(metrics)


def calculate_v_x(dvh, x, label=None):
    """Get the volume (in cc) which receives x dose

    Args:
        dvh (pandas.DataFrame): DVH DataFrame as produced by calculate_dvh_for_labels
        x (float|list): The dose threshold (or list of dose thresholds) to get the volume for.
        label (str, optional): The label to compute the metric for. Computes for all metrics if not
            set. Defaults to None.

    Returns:
        pandas.DataFrame: Data frame with a row for each label containing the metric and value.
    """

    if label:
        dvh = dvh[dvh.label == label]

    if not isinstance(x, list):
        x = [x]

    bins = np.array([b for b in dvh.columns if isinstance(b, float)])
    values = np.array(dvh[bins])

    metrics = []
    for idx in range(len(dvh)):
        d = dvh.iloc[idx]

        m = {"label": d.label}

        for threshold in x:
            value = np.interp(threshold, bins, values[idx]) * d.cc

            metric_name = f"V{threshold}"
            if threshold - int(threshold) == 0:
                metric_name = f"V{int(threshold)}"

            m[metric_name] = value

        metrics.append(m)

    return pd.DataFrame(metrics)


def calculate_d_cc_x(dvh, x, label=None):
    """Compute the dose which is received by cc of the volume

    Args:
        dvh (pandas.DataFrame): DVH DataFrame as produced by calculate_dvh_for_labels
        x (float|list): The cc (or list of cc's) to compute the dose at.
        label (str, optional): The label to compute the metric for. Computes for all metrics if not
            set. Defaults to None.

    Returns:
        pandas.DataFrame: Data frame with a row for each label containing the metric and value.
    """

    if label:
        dvh = dvh[dvh.label == label]

    if not isinstance(x, list):
        x = [x]

    metrics = []
    for idx in range(len(dvh)):

        d = dvh.iloc[idx]
        m = {"label": d.label}

        for threshold in x:
            cc_at = (threshold / dvh[dvh.label == d.label].cc.iloc[0]) * 100
            cc_at = min(cc_at, 100)
            cc_val = calculate_d_x(dvh[dvh.label == d.label], cc_at)[f"D{cc_at}"].iloc[0]

            m[f"D{threshold}cc"] = cc_val

        metrics.append(m)

    return pd.DataFrame(metrics)



import os
import tempfile
import shutil
import urllib.request
from pathlib import Path
import logging
import SimpleITK as sitk


logger = logging.getLogger(__name__)


def available_nnunet_models():
    """Fetch the available nnUNet models

    Raises:
        ImportError: Raised when nnUNet library hasn't been installed.

    Returns:
        dict: Dictionary describing models available.
    """

    try:
        from nnunet.inference.pretrained_models.download_pretrained_model import (
            get_available_models,
        )
    except ImportError:
        raise ImportError(
            "nnUNet Library not found. Be sure to install platipy with the required extras: 'pip install platipy[cardiac]' or 'pip install platipy[nnunet]'"
        )

    available_models = get_available_models()
    available_models["Task400_OPEN_HEART_1FOLD"] = {
        "description": "Whole heart model (all folds, 3d_lowres only) trained on data from"
        "TCIA (NSCLC-Radiomics & LCTSC)",
        "url": "https://zenodo.org/record/6585664/files/Task400_OPEN_HEART_3d_lowres.zip?download=1",
    }
    return available_models


NNUNET_SETTINGS_DEFAULTS = {
    "task": "Task400_OPEN_HEART_1FOLD",
    "trainer_class_name": "nnUNetTrainerV2",
    "cascade_trainer_class_name": "nnUNetTrainerV2CascadeFullRes",
    "folds": None,
    "model": "3d_lowres",
    "lowres_segmentations": None,
    "num_threads_preprocessing": 6,
    "num_threads_nifti_save": 2,
    "disable_tta": False,
    "all_in_gpu": None,
    "disable_mixed_precision": False,
    "chk": "model_final_checkpoint",
}


def setup_nnunet_environment():
    """Inserts suitable location for nnUNet environemnt variables if they are not already set in
    the users environment.
    """

    # Setup the nnUNet environment variables
    if not "RESULTS_FOLDER" in os.environ:
        home = Path.home()
        platipy_dir = home.joinpath(".platipy")
        home.mkdir(exist_ok=True)
        os.environ["RESULTS_FOLDER"] = str(platipy_dir.joinpath("nnUNet_models"))

        # Don't really need these here but set them anyway to supress warnings
        os.environ["nnUNet_raw_data_base"] = tempfile.mkdtemp()
        os.environ["nnUNet_preprocessed"] = tempfile.mkdtemp()


def download_and_install_nnunet_task(task_name, zip_url):
    """Downloads the Zip file and then installs via nnUNet.

    Avoid using the nnUNet built in function since it doesn't work when a HTTP_PROXY/HTTPS_PROXY
    are set.

    Args:
        task_name (str): Task ID and name
        zip_url (str): Zip file URL
    """

    try:
        from nnunet.inference.pretrained_models.download_pretrained_model import (
            install_model_from_zip_file,
        )
    except ImportError:
        raise ImportError(
            "nnUNet Library not found. Be sure to install platipy with the required extras: 'pip install platipy[cardiac]' or 'pip install platipy[nnunet]'"
        )

    logger.info("Installing Task %s from %s", task_name, zip_url)
    temp_dir = tempfile.mkdtemp()
    temp_file = Path(temp_dir).joinpath(f"{task_name}.zip")
    with urllib.request.urlopen(zip_url) as dl_file:
        with open(temp_file, "wb") as out_file:
            out_file.write(dl_file.read())

    install_model_from_zip_file(temp_file)
    shutil.rmtree(temp_dir)


def run_segmentation(img, settings=NNUNET_SETTINGS_DEFAULTS):

    setup_nnunet_environment()

    # Import in here to make sure environment is already set
    try:
        from nnunet.inference.predict import predict_from_folder
    except ImportError:
        logger.error(
            "nnUNet is not installed. Please pip install nnunet to use this functionality"
        )

    nnunet_model_path = Path(os.environ["RESULTS_FOLDER"])

    task = settings["task"]
    model = settings["model"]

    # Check if task model is already installed
    task_path = nnunet_model_path.joinpath("nnUNet", model, task)
    print(task_path)

    if not task_path.exists():
        # Check if the task is available and install it if so
        available_models = available_nnunet_models()

        if not task in available_models:
            raise ValueError(f"{task} not available")

        task_url = available_models[task]["url"]
        download_and_install_nnunet_task(task, task_url)

    # Prepare the image in a temporary directory for nnunet to run on
    input_path = Path(tempfile.mkdtemp())
    io_path = input_path.joinpath(f"{settings['task']}_0000.nii.gz")
    sitk.WriteImage(img, str(io_path))

    output_path = Path(tempfile.mkdtemp())

    model = settings["model"]
    folds = settings["folds"]
    num_threads_preprocessing = settings["num_threads_preprocessing"]
    num_threads_nifti_save = settings["num_threads_nifti_save"]
    lowres_segmentations = settings["lowres_segmentations"]
    all_in_gpu = settings["all_in_gpu"]
    disable_mixed_precision = settings["disable_mixed_precision"]
    disable_tta = settings["disable_tta"]
    trainer_class_name = settings["trainer_class_name"]
    cascade_trainer_class_name = settings["cascade_trainer_class_name"]
    mode = "normal"
    default_plans_identifier = "nnUNetPlansv2.1"
    chk = settings["chk"]

    if model == "3d_cascade_fullres":
        trainer = cascade_trainer_class_name
    else:
        trainer = trainer_class_name

    model_folder_name = task_path.joinpath(trainer + f"__{default_plans_identifier}")

    predict_from_folder(
        str(model_folder_name),
        str(input_path),
        str(output_path),
        folds,
        False,
        num_threads_preprocessing,
        num_threads_nifti_save,
        lowres_segmentations,
        0,
        1,
        not disable_tta,
        overwrite_existing=True,
        mode=mode,
        overwrite_all_in_gpu=all_in_gpu,
        mixed_precision=not disable_mixed_precision,
        step_size=0.5,
        checkpoint_name=chk,
    )

    results = {}
    for op in output_path.glob("*.nii.gz"):

        label_map = sitk.ReadImage(str(op))
        num_labels = sitk.GetArrayViewFromImage(label_map).max()

        for l in range(num_labels):
            results[f"Struct_{l}"] = label_map == (l + 1)

    os.remove(io_path)

    return results



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import SimpleITK as sitk


def vector_angle(v1, v2, smallest=True):
    """Return the angle between two vectors

    Args:
        v1 (np.array): A three-dimensional vector
        v2 (np.array): A three-dimensional vector
        smallest (bool, optional): If True, the angle is the smallest (i.e. ignoring direction).
            If False, direction is taken into account, so angle can be obtuse. Defaults to True.

    Returns:
        float: The angle in radians
    """
    v1 = np.array(v1)
    v2 = np.array(v2)
    v1_norm = v1 / np.linalg.norm(v1)
    v2_norm = v2 / np.linalg.norm(v2)
    dot_product = np.dot(v1_norm, v2_norm)
    if smallest:
        dot_product = np.abs(dot_product)
    angle = np.arccos(dot_product)
    return angle


def rotate_image(
    img,
    rotation_centre=(0, 0, 0),
    rotation_axis=(1, 0, 0),
    rotation_angle_radians=0,
    interpolation=sitk.sitkNearestNeighbor,
    default_value=0,
):
    """Rotates an image

    Args:
        img (SimpleITK.Image): The image to rotate
        rotation_centre (tuple, optional): The centre of rotation (in physical coordinates).
            Defaults to (0, 0, 0).
        rotation_axis (tuple, optional): The axis of rotation. Defaults to (1, 0, 0).
        rotation_angle_radians (float, optional): The angle of rotation. Defaults to 0.
        interpolation (int, optional): Final interpolation. Defaults to sitk.sitkNearestNeighbor.
        default_value (int, optional): Default value. Defaults to 0.

    Returns:
        SimpleITK.Image: The rotated image, resampled into the original space.
    """

    # Define the transform, using predefined centre of rotation and given angle
    rotation_transform = sitk.VersorRigid3DTransform()
    rotation_transform.SetCenter(rotation_centre)
    rotation_transform.SetRotation(rotation_axis, rotation_angle_radians)

    # Resample the image using the rotation transform
    resampled_image = sitk.Resample(
        img,
        rotation_transform,
        interpolation,
        default_value,
        img.GetPixelID(),
    )

    return resampled_image



# Copyright 2020 University of New South Wales, University of Sydney

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This script was provided by Jason Dowling and written by Parnesh Raniga, CSIRO, Brisbane.


from pathlib import Path
import SimpleITK as sitk
import numpy as np
import logging
from matplotlib import cm

from rt_utils import RTStructBuilder

logger = logging.getLogger(__name__)


def convert_nifti(dcm_path, mask_input, output_file, color_map=cm.get_cmap("rainbow")):
    """Convert a set of masks to a DICOM RTStruct object.

    This function now utilises the rt-utils package: https://github.com/qurit/rt-utils
    We keep this function in here for compatibility

    Args:
        dcm_path (str|pathlib.Path): Path to the reference DICOM series
        mask_input (dict|list): A dictionary containing the name as key and image as value. Or a
                                list of string with comma separated name and mask paths (name,path)
        output_file (str|pathlib.Path): The path to the file to write the RTStruct
         color_map (matplotlib.colors.Colormap, optional): Colormap to use for output. Defaults to
                                                           cm.get_cmap("rainbow").
    """

    logger.info("Will convert the following Nifti masks to RTStruct:")

    masks = {}  # Dict stores key value pairs for masks
    if isinstance(mask_input, dict):
        # Dict was already passed in
        masks = mask_input
    else:
        # Otherwise convert list of comma separated masks
        for mask in mask_input:
            mask_parts = mask.split(",")
            masks[mask_parts[0]] = mask_parts[1]

    if not isinstance(dcm_path, Path):
        dcm_path = Path(dcm_path)

    dcm_series_path = None
    if dcm_path.is_file():
        dcm_series_path = dcm_path.parent
    else:
        dcm_series_path = dcm_path

    rtstruct = RTStructBuilder.create_new(dicom_series_path=str(dcm_series_path))

    for mask_name in masks:

        # Use a hash of the name to get the color from the supplied color map
        color = color_map(hash(mask_name) % 256)
        color = color[:3]
        color = [int(c * 255) for c in color]

        mask = masks[mask_name]
        if not isinstance(mask, sitk.Image):
            mask = sitk.ReadImage(str(mask))

        bool_arr = sitk.GetArrayFromImage(mask) != 0
        bool_arr = np.transpose(bool_arr, (1, 2, 0))
        rtstruct.add_roi(mask=bool_arr, color=color, name=mask_name)

    rtstruct.save(str(output_file))



#!/usr/bin/env python

# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import logging

logger = logging.getLogger(__name__)

from platipy.cli import (
    dicom_crawler,
    segmentation,
    nifti_to_rtstruct,
    rtstruct_to_nifti,
    nifti_to_series,
    tcia_download,
)

tools = {
    "dicom_crawler": dicom_crawler.click_command,
    "segmentation": segmentation.click_command,
    "nifti_to_rtstruct": nifti_to_rtstruct.click_command,
    "rtstruct_to_nifti": rtstruct_to_nifti.click_command,
    "nifti_to_series": nifti_to_series.click_command,
    "tcia-download": tcia_download.click_command,
}

# If backend tools are installed, then provide manage tools
try:
    from platipy.backend.manage import cli

    tools["manage"] = cli
except ImportError:
    logger.info(
        "PlatiPy Backend requirements not installed. Install the requirements listed "
        "at: https://github.com/pyplati/platipy/blob/master/requirements-backend.txt to use "
        "backend service functionality"
    )


def platipy_cli():
    """Run the PlatiPy Command Line Interface"""
    if len(sys.argv) == 1 or not sys.argv[1] in tools:
        print("")
        print("  PlatiPy CLI (Command Line Interface)")
        print("  ------------------------------------")
        print("")
        print("  Usage: platipy [tool]")
        print("")
        print("  Supply the name of the desired tool:")
        for key in tools:
            print(f"    {key}")
        print("")
        sys.exit()

    tool = sys.argv[1]
    del sys.argv[1]  # Remove the tool from the arg list so that click can run the command properly
    tools[tool]()


if __name__ == "__main__":
    platipy_cli()



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from platipy.backend import db
import datetime
import json

from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import DeclarativeMeta

# This custom encoder takes care of converting our SQLAlchemy models to a JSON
# encodable format.
class AlchemyEncoder(json.JSONEncoder):
    def default(self, obj):

        if isinstance(obj.__class__, DeclarativeMeta):
            # an SQLAlchemy class
            fields = {}
            for field in [
                x for x in dir(obj) if not x.startswith("_") and x != "metadata"
            ]:
                data = obj.__getattribute__(field)

                try:
                    # this will fail on non-encodable values, like other classes
                    json.dumps(data)
                    fields[field] = data
                except TypeError:
                    if isinstance(data, datetime.datetime):
                        r = data.isoformat()

                        if data.microsecond:
                            r = r[:23] + r[26:]
                        if r.endswith("+00:00"):
                            r = r[:-6] + "Z"
                        fields[field] = r
                    elif isinstance(data, datetime.date):
                        fields[field] = data.isoformat()
                    elif isinstance(data, datetime.time):
                        if is_aware(data):
                            raise ValueError(
                                "JSON can't represent timezone-aware times."
                            )
                        r = data.isoformat()
                        if data.microsecond:
                            r = r[:12]
                        fields[field] = r
                    elif isinstance(data, DicomLocation):
                        fields[field] = self.default(data)
                    elif isinstance(data, list):
                        o = []
                        for d in data:
                            o.append(self.default(d))
                        fields[field] = o

            # a json-encodable dict
            return fields

        return json.JSONEncoder.default(self, obj)


def default_timeout():

    return datetime.datetime.utcnow() + datetime.timedelta(hours=24)


class APIKey(db.Model):
    __tablename__ = "APIKey"

    key = db.Column(db.String(80), primary_key=True)
    name = db.Column(db.String(80))

    is_admin = db.Column(db.Boolean, default=False)

    def __repr__(self):
        return "{0}: {1} (Admin: {2})".format(self.name, self.key, self.is_admin)


class DicomLocation(db.Model):
    __tablename__ = "DicomLocation"

    id = db.Column(db.Integer, primary_key=True)

    name = db.Column(db.String(128))

    host = db.Column(db.String(128), nullable=False)
    port = db.Column(db.Integer, nullable=False)
    ae_title = db.Column(db.String(128))

    owner_key = db.Column(db.String(80), db.ForeignKey("APIKey.key"), nullable=False)

    def __repr__(self):
        return "{0} {1} {2}".format(self.host, self.port, self.ae_title)


class Dataset(db.Model):
    __tablename__ = "Dataset"

    id = db.Column(db.Integer, primary_key=True)
    owner_key = db.Column(db.String(80), db.ForeignKey("APIKey.key"), nullable=False)

    input_data_objects = relationship(
        "DataObject",
        primaryjoin="and_(DataObject.dataset_id == Dataset.id, DataObject.is_input == True)",
    )
    output_data_objects = relationship(
        "DataObject",
        primaryjoin="and_(DataObject.dataset_id == Dataset.id, DataObject.is_input == False)",
        overlaps="input_data_objects"
    )

    # The Dicom location from which to retrieve data
    from_dicom_location_id = db.Column(db.Integer, db.ForeignKey("DicomLocation.id"))
    from_dicom_location = relationship(
        "DicomLocation", foreign_keys=[from_dicom_location_id]
    )

    # The Dicom location to send data to
    to_dicom_location_id = db.Column(db.Integer, db.ForeignKey("DicomLocation.id"))
    to_dicom_location = relationship(
        "DicomLocation", foreign_keys=[to_dicom_location_id]
    )

    timestamp = db.Column(db.DateTime, nullable=False, default=datetime.datetime.utcnow)
    timeout = db.Column(db.DateTime, nullable=False, default=default_timeout)

    def __repr__(self):
        return "{0}: {1}".format(self.id, self.timestamp)


class DataObject(db.Model):
    __tablename__ = "DataObject"

    id = db.Column(db.Integer, primary_key=True)
    dataset_id = db.Column(db.Integer, db.ForeignKey("Dataset.id"), nullable=False)
    dataset = relationship("Dataset", overlaps="input_data_objects,output_data_objects")
    is_input = db.Column(db.Boolean, default=False)

    path = db.Column(db.String(256))

    type = db.Column(db.String(32), nullable=False, default="FILE")

    series_instance_uid = db.Column(db.String(256))

    meta_data = db.Column(db.JSON)
    status = db.Column(db.JSON)

    is_fetched = db.Column(db.Boolean, default=False)
    is_sent = db.Column(db.Boolean, default=False)

    timestamp = db.Column(db.DateTime, nullable=False, default=datetime.datetime.utcnow)

    parent_id = db.Column(db.Integer, db.ForeignKey("DataObject.id"), index=True)
    parent = relationship("DataObject", remote_side=[id])
    children = relationship(
        "DataObject", primaryjoin="and_(DataObject.parent_id == DataObject.id)", overlaps="parent"
    )

    def __repr__(self):
        return "{0} - {1}: {2}".format(self.dataset_id, self.id, self.timestamp)



# Copyright 2020 CSIRO, University of New South Wales, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Created on Wed Feb 21 16:39:27 2018

Bronchus segmentation.  The superior extent of bronchus is around 4cm SUP from Carina (From first
slice where two airways become visible).
In this code I'm using 2cm as it's easier to detect where the airways are getting wider

Areas to improve: parameters could be improved (eg. the median filter, carina detection, etc). The
GenLung_mask is based on old ITK code from a masters student.  I think we can replace this function
by checking the top (sup) slice for an airhole and then connected thresholding.

Code fails on two Liverpool cases:  13 (lungs appear in the sup slice) and 36 (the mask failed to
generate - need to look at this)

@author: Jason Dowling (CSIRO)
"""

import SimpleITK as sitk

from platipy.imaging.utils.lung import detect_holes, get_lung_mask


def fast_mask(img, start, end):
    """Fast masking for area of a 3D volume .

    SimpleITK lacks iterators so voxels need to be set with SetPixel which is horribly slow.
    This code uses numpy arrays to reduce time for one volume from around one minute to about 0.5s

    Args:
        image: Input 3D binary image, start slice for masking (value=0), end slice for masking

    Returns:
        Masked image.  This may be in float, so it might need casting back to original pixel type.
    """
    np_img = sitk.GetArrayFromImage(img).astype(float)
    np_img[start:end, :, :] = 0
    new_img = sitk.GetImageFromArray(np_img)
    new_img.SetSpacing(img.GetSpacing())
    new_img.SetOrigin(img.GetOrigin())
    new_img.SetDirection(img.GetDirection())
    return new_img


def get_distance(a_mask, b_mask):
    """Get the nearest distance between two masks.

    Args:
        a_mask: The first mask
        b_mask: The second mask
        dest: Working directory to output intermediate files

    Returns:
        None
    """

    # load lung mask from previous step
    try:
        a_mask = sitk.ReadImage(a_mask)
    except:
        print("File read failed " + a_mask)
        raise

    try:
        b_mask = sitk.ReadImage(b_mask)
    except:
        print("File read failed " + b_mask)
        raise

    # 1. Generate distance map from surface of Contour1
    smdm = sitk.SignedMaurerDistanceMapImageFilter()
    smdm.UseImageSpacingOn()
    smdm.InsideIsPositiveOff()
    smdm.SquaredDistanceOff()
    result = smdm.Execute(a_mask)

    # Subtract 1 from the mask, making 0's -> -1 and 1's -> 0.
    b_mask = b_mask - 1

    # Multiply this result by -10000, making -1's -> 10000.
    # It is assumed that 10000mm > than the max distance
    b_mask = b_mask * -10000

    result = result + sitk.Cast(b_mask, result.GetPixelIDValue())

    sif = sitk.StatisticsImageFilter()
    sif.Execute(result)
    dist = sif.GetMinimum()

    return dist


def generate_lung_mask(img):
    """Generate initial airway mask (includes lungs).

    Args:
        img: The SimpleITK CT image to segment the lungs in

    Returns:
        lung_mask: The mask containing the lung segmentation
    """

    print("Generating Lung Mask...")

    (label_image, labels) = detect_holes(img)
    lung_mask = get_lung_mask(label_image, labels)

    print("Generating Lung Mask... Done")

    return lung_mask


default_settings = {
    "fast_mode": True,
    "extend_from_carina_mm": 40,
    "minimum_tree_half_physical_size": 1000,
    "lung_mask_hu_values": [-750, -775, -800, -825, -850, -900, -700, -950, -650],
    "distance_from_supu_slice_values": [3, 10, 20],
    "expected_physical_size_range": [22000, 150000],
}


def generate_airway_mask(dest, img, lung_mask, config_dict=None):
    """Generate final bronchus segmentation .

    Args:
        image: path for CT image and mask destination

    Returns:
        None
    """

    if not config_dict:
        config_dict = default_settings

    fast_mode = config_dict["fast_mode"]
    extend_from_carina_mm = config_dict["extend_from_carina_mm"]
    lung_mask_hu_values = config_dict["lung_mask_hu_values"]
    minimum_tree_half_physical_size = config_dict["minimum_tree_half_physical_size"]
    distance_from_supu_slice_values = config_dict["distance_from_supu_slice_values"]
    expected_physical_size_range = config_dict["expected_physical_size_range"]

    z_size = img.GetDepth()
    z_spacing = img.GetSpacing()[2]
    extend_from_carina = round(extend_from_carina_mm / z_spacing)

    # Identify airway start on superior slice
    label_shape = sitk.LabelIntensityStatisticsImageFilter()
    connected_component = sitk.ConnectedComponentImageFilter()
    connected_threshold_filter = sitk.ConnectedThresholdImageFilter()

    print("-------------------------------")

    label_shape.Execute(lung_mask, img)
    for label in label_shape.GetLabels():
        print(
            "summary of lung intensities.  Mean: "
            + str(label_shape.GetMean(label))
            + " sd: "
            + str(label_shape.GetStandardDeviation(label))
            + " median: "
            + str(label_shape.GetMedian(label))
        )

    loop_count = 0
    processed_correctly = False

    best_result = None
    best_result_sim = 0
    best_lung_mask_hu = 0
    best_distance_from_sup_slice = 0

    for k in range(2):

        if processed_correctly and fast_mode:
            break

        if k == 1:
            # Try smoothing the lung mask - effects all tests below
            median_filter = sitk.MedianImageFilter()
            median_filter.SetRadius(1)
            lung_mask = median_filter.Execute(lung_mask)

        for distance_from_sup_slice in distance_from_supu_slice_values:

            if processed_correctly and fast_mode:
                break

            label_slice = lung_mask[
                :,
                :,
                z_size - distance_from_sup_slice - 10 : z_size - distance_from_sup_slice,
            ]  # works for both cases 22 and 17
            img_slice = img[
                :,
                :,
                z_size - distance_from_sup_slice - 10 : z_size - distance_from_sup_slice,
            ]

            connected = connected_component.Execute(label_slice)
            # In case there are multiple air regions at the top, select the region with
            # the max elongation as the airway seed
            # Also check that the region has a minimum physical size, to filter out other
            # unexpected regions
            max_elong = 0
            airway_open = [0, 0, 0]
            label_shape.Execute(connected, img_slice)
            for label in label_shape.GetLabels():
                if (
                    label_shape.GetElongation(label) > max_elong
                    and label_shape.GetPhysicalSize(label) > 2000
                ):
                    centre = img.TransformPhysicalPointToIndex(label_shape.GetCentroid(label))
                    max_elong = label_shape.GetElongation(label)
                    airway_open = [int(centre[0]), int(centre[1]), int(centre[2])]

            # just check the opening is at the right location
            centroid_mask_val = lung_mask.GetPixel(airway_open[0], airway_open[1], airway_open[2])

            if centroid_mask_val == 0:
                print(
                    """Error locating trachea centroid.  Usually because of additional air features
                    on this slice"""
                )

                # No point in doing the airway segmentation here as airway opening wasn't detected
                continue

            print("*Airway opening: " + str(airway_open))
            print(
                "*Voxel HU at opening: "
                + str(lung_mask.GetPixel(airway_open[0], airway_open[1], airway_open[2]))
            )

            for lung_mask_hu in lung_mask_hu_values:

                print("--------------------------------------------")
                print("Extracting airways.  Iteration: " + str(loop_count))
                print("*Lung Mask HU: " + str(lung_mask_hu))
                print("*Slices from sup for airway opening: " + str(distance_from_sup_slice))
                if k == 1:
                    print("*Mask median smoothing on")
                loop_count += 1

                connected_threshold_filter.SetSeedList([airway_open])
                connected_threshold_filter.SetLower(-2000)
                connected_threshold_filter.SetUpper(lung_mask_hu)
                result = connected_threshold_filter.Execute(img)

                writer = sitk.ImageFileWriter()
                writer.SetFileName(dest + "/airwaysMask.nii.gz")
                writer.Execute(result)

                # Dilate and check if the output is in the expected range
                binary_dilate = sitk.BinaryDilateImageFilter()
                binary_dilate.SetKernelRadius(2)
                result = binary_dilate.Execute(result)

                # check size of label - if it's too large the lungs have been included..
                result = sitk.Cast(result, lung_mask.GetPixelIDValue())
                label_shape.Execute(result, img)
                airway_mask_physical_size = -1
                for label in label_shape.GetLabels():
                    airway_mask_physical_size = int(label_shape.GetPhysicalSize(label))
                    roundness = float(label_shape.GetRoundness(label))
                    elongation = float(label_shape.GetElongation(label))

                this_processed_correctly = False
                if airway_mask_physical_size < 0:
                    print("No labels found in mask")
                    continue
                elif airway_mask_physical_size > expected_physical_size_range[1]:
                    print(
                        " Airway Mask size failed (> "
                        + str(expected_physical_size_range[1])
                        + "): "
                        + str(airway_mask_physical_size)
                    )
                elif airway_mask_physical_size < expected_physical_size_range[0]:
                    print(
                        " Airway Mask size failed (< "
                        + str(expected_physical_size_range[0])
                        + "): "
                        + str(airway_mask_physical_size)
                    )
                else:
                    print(" Airway Mask size passed: " + str(airway_mask_physical_size))
                    processed_correctly = True
                    this_processed_correctly = True

                print(" Roundness: " + str(roundness))
                print(" Elongation: " + str(elongation))
                # target_size = (
                #     expected_physical_size_range[1] + expected_physical_size_range[0]
                # ) / 2
                # size_sim = abs(airway_mask_physical_size - target_size)

                if airway_mask_physical_size > best_result_sim and this_processed_correctly:
                    best_result_sim = airway_mask_physical_size
                    best_result = result
                    best_lung_mask_hu = lung_mask_hu
                    best_distance_from_sup_slice = distance_from_sup_slice

    if not processed_correctly:
        print(" Unable to process correctly!!!")

    print("Selected Lung Mask HU: " + str(best_lung_mask_hu))

    # process in 2D - check label elongation and size.
    corina_slice = -1
    lssif = sitk.LabelShapeStatisticsImageFilter()
    for idx_slice in range(z_size - best_distance_from_sup_slice, 0, -1):

        cut_mask = fast_mask(best_result, idx_slice, z_size)
        cut_mask = sitk.Cast(cut_mask, lung_mask.GetPixelIDValue())

        label_image = connected_component.Execute(cut_mask)

        num_regions = connected_component.GetObjectCount()

        if num_regions == 2:
            lssif.Execute(label_image)

            phys_size_0 = int(lssif.GetPhysicalSize(1))
            phys_size_1 = int(lssif.GetPhysicalSize(2))

            if (
                phys_size_0 > minimum_tree_half_physical_size
                and phys_size_1 > minimum_tree_half_physical_size
            ):

                corina_slice = idx_slice
                break

    if corina_slice >= 0:
        print(f" Cropping from slice: {corina_slice} + {extend_from_carina} slices")
        best_result = fast_mask(best_result, corina_slice + extend_from_carina, z_size)

    best_result = sitk.Cast(best_result, lung_mask.GetPixelIDValue())

    return best_result



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import shutil
import tempfile
import zipfile
from pathlib import Path

import SimpleITK as sitk
import numpy as np

from platipy.imaging.registration.utils import apply_transform, convert_mask_to_reg_structure

from platipy.imaging.registration.linear import (
    linear_registration,
)

from platipy.imaging.registration.deformable import (
    fast_symmetric_forces_demons_registration,
)

from platipy.imaging.label.fusion import (
    process_probability_image,
    compute_weight_map,
    combine_labels,
)
from platipy.imaging.label.iar import run_iar
from platipy.imaging.utils.vessel import vessel_spline_generation

from platipy.imaging.utils.valve import (
    generate_valve_from_great_vessel,
    generate_valve_using_cylinder,
)

from platipy.imaging.utils.conduction import (
    geometric_sinoatrialnode,
    geometric_atrioventricularnode,
)

from platipy.imaging.utils.crop import label_to_roi, crop_to_roi
from platipy.imaging.generation.mask import extend_mask
from platipy.imaging.label.utils import binary_encode_structure_list, correct_volume_overlap
from platipy.imaging.projects.nnunet.run import (
    run_segmentation,
    available_nnunet_models,
    setup_nnunet_environment,
    NNUNET_SETTINGS_DEFAULTS,
)
from platipy.utils import download_and_extract_zip_file

logger = logging.getLogger(__name__)

ATLAS_PATH = "/atlas"
if "ATLAS_PATH" in os.environ:
    ATLAS_PATH = os.environ["ATLAS_PATH"]
else:
    home = Path.home()
    platipy_dir = home.joinpath(".platipy")
    platipy_dir.mkdir(exist_ok=True)
    ATLAS_PATH = str(platipy_dir.joinpath("cardiac", "test_atlas"))

CARDIAC_SETTINGS_DEFAULTS = {
    "atlas_settings": {
        "atlas_id_list": [
            "03",
            "05",
            "08",
            "10",
            "11",
            "12",
            "13",
            "16",
            "24",
            "35",
        ],
        "atlas_structure_list": [
            "AORTICVALVE",
            "ASCENDINGAORTA",
            "LANTDESCARTERY",
            "LCIRCUMFLEXARTERY",
            "LCORONARYARTERY",
            "LEFTATRIUM",
            "LEFTVENTRICLE",
            "MITRALVALVE",
            "PULMONARYARTERY",
            "PULMONICVALVE",
            "RCORONARYARTERY",
            "RIGHTATRIUM",
            "RIGHTVENTRICLE",
            "SVC",
            "TRICUSPIDVALVE",
            "WHOLEHEART",
        ],
        "atlas_path": ATLAS_PATH,
        "atlas_image_format": "Case_{0}/Images/Case_{0}_CROP.nii.gz",
        "atlas_label_format": "Case_{0}/Structures/Case_{0}_{1}_CROP.nii.gz",
        "crop_atlas_to_structures": False,
        "crop_atlas_expansion_mm": (20, 20, 40),
        "guide_structure_name": "WHOLEHEART",
        "superior_extension": 30,
    },
    "auto_crop_target_image_settings": {
        "expansion_mm": [20, 20, 40],
    },
    "linear_registration_settings": {
        "reg_method": "affine",
        "shrink_factors": [16, 8, 4],
        "smooth_sigmas": [0, 0, 0],
        "sampling_rate": 0.75,
        "default_value": -1000,
        "number_of_iterations": 50,
        "metric": "mean_squares",
        "optimiser": "gradient_descent_line_search",
        "verbose": False,
    },
    "structure_guided_registration_settings": {
        "isotropic_resample": True,
        "resolution_staging": [
            16,
            8,
            2,
        ],  # specify voxel size (mm) since isotropic_resample is set
        "iteration_staging": [50, 50, 50],
        "smoothing_sigmas": [0, 0, 0],
        "ncores": 8,
        "default_value": 0,
        "verbose": False,
    },
    "deformable_registration_settings": {
        "isotropic_resample": True,
        "resolution_staging": [
            6,
            3,
            1.5,
        ],  # specify voxel size (mm) since isotropic_resample is set
        "iteration_staging": [200, 150, 100],
        "smoothing_sigmas": [0, 0, 0],
        "ncores": 8,
        "default_value": 0,
        "verbose": False,
    },
    "iar_settings": {
        "reference_structure": False,
        "smooth_distance_maps": True,
        "smooth_sigma": 1,
        "z_score_statistic": "mad",
        "outlier_method": "iqr",
        "outlier_factor": 1.5,
        "min_best_atlases": 5,
        "project_on_sphere": False,
    },
    "label_fusion_settings": {
        "vote_type": "unweighted",
        "vote_params": None,
        "optimal_threshold": {
            "AORTICVALVE": 0.5,
            "ASCENDINGAORTA": 0.44,
            "LEFTATRIUM": 0.40,
            "LEFTVENTRICLE": 0.45,
            "MITRALVALVE": 0.5,
            "PULMONARYARTERY": 0.46,
            "PULMONICVALVE": 0.5,
            "RIGHTATRIUM": 0.38,
            "RIGHTVENTRICLE": 0.42,
            "SVC": 0.44,
            "TRICUSPIDVALVE": 0.5,
            "WHOLEHEART": 0.5,
        },
    },
    "vessel_spline_settings": {
        "vessel_name_list": [
            "LANTDESCARTERY",
            "LCIRCUMFLEXARTERY",
            "LCORONARYARTERY",
            "RCORONARYARTERY",
        ],
        "vessel_radius_mm_dict": {
            "LANTDESCARTERY": 2,
            "LCIRCUMFLEXARTERY": 2,
            "LCORONARYARTERY": 2,
            "RCORONARYARTERY": 2,
        },
        "scan_direction_dict": {
            "LANTDESCARTERY": "z",
            "LCIRCUMFLEXARTERY": "z",
            "LCORONARYARTERY": "x",
            "RCORONARYARTERY": "z",
        },
        "stop_condition_type_dict": {
            "LANTDESCARTERY": "count",
            "LCIRCUMFLEXARTERY": "count",
            "LCORONARYARTERY": "count",
            "RCORONARYARTERY": "count",
        },
        "stop_condition_value_dict": {
            "LANTDESCARTERY": 2,
            "LCIRCUMFLEXARTERY": 2,
            "LCORONARYARTERY": 2,
            "RCORONARYARTERY": 2,
        },
    },
    "geometric_segmentation_settings": {
        "run_geometric_algorithms": True,
        "geometric_name_suffix": "_GEOMETRIC",
        "atlas_structure_names": {
            "atlas_left_ventricle": "LEFTVENTRICLE",
            "atlas_right_ventricle": "RIGHTVENTRICLE",
            "atlas_left_atrium": "LEFTATRIUM",
            "atlas_right_atrium": "RIGHTATRIUM",
            "atlas_ascending_aorta": "ASCENDINGAORTA",
            "atlas_pulmonary_artery": "PULMONARYARTERY",
            "atlas_superior_vena_cava": "SVC",
            "atlas_whole_heart": "WHOLEHEART",
        },
        "valve_definitions": {
            "mitral_valve_thickness_mm": 10,
            "mitral_valve_radius_mm": 15,
            "tricuspid_valve_thickness_mm": 10,
            "tricuspid_valve_radius_mm": 15,
            "pulmonic_valve_thickness_mm": 10,
            "aortic_valve_thickness_mm": 10,
        },
        "conduction_system_definitions": {
            "sinoatrial_node_radius_mm": 10,
            "atrioventricular_node_radius_mm": 10,
        },
    },
    "postprocessing_settings": {
        "run_postprocessing": True,
        "binaryfillhole_mm": 3,
        "structures_for_binaryfillhole": [
            "ASCENDINGAORTA",
            "LEFTATRIUM",
            "LEFTVENTRICLE",
            "RIGHTATRIUM",
            "RIGHTVENTRICLE",
            "SVC",
            "AORTICVALVE",
            "MITRALVALVE",
            "PULMONICVALVE",
            "TRICUSPIDVALVE",
            "WHOLEHEART",
        ],
        "structures_for_overlap_correction": [
            "ASCENDINGAORTA",
            "LEFTATRIUM",
            "LEFTVENTRICLE",
            "RIGHTATRIUM",
            "RIGHTVENTRICLE",
            "PULMONARYARTERY",
            "SVC",
        ],
    },
    "return_atlas_guide_structure": False,
    "return_as_cropped": False,
    "return_proba_as_contours": False,
}

OPEN_ATLAS_URL = "https://zenodo.org/record/6592437/files/open_atlas.zip?download=1"

OPEN_ATLAS_SETTINGS = CARDIAC_SETTINGS_DEFAULTS.copy()
OPEN_ATLAS_SETTINGS["atlas_settings"] = {
    "atlas_id_list": [
        "LCTSC-Test-S2-201",
        "LCTSC-Test-S2-203",
        "LCTSC-Test-S3-201",
        #        "LCTSC-Train-S1-007",
        #        "LCTSC-Train-S2-005",
        #        "LCTSC-Train-S3-004",
        "LUNG1-002",
        "LUNG1-009",
        "LUNG1-021",
        #        "LUNG1-027",
        #        "LUNG1-035",
        #        "LUNG1-037",
        #        "LUNG1-055",
        "LUNG1-067",
        #        "LUNG1-074",
        #        "LUNG1-076",
        #        "LUNG1-092",
        #        "LUNG1-143",
        #        "LUNG1-210",
        "LUNG1-226",
    ],
    "atlas_structure_list": [
        "Atrium_L",
        "Ventricle_L",
        "Atrium_R",
        "Ventricle_R",
        "A_Aorta",
        "A_Pulmonary",
        "V_Venacava_S",
        "A_LAD",
        "A_Coronary_L",
        "A_Cflx",
        "A_Coronary_R",
        "Heart",
    ],
    "atlas_path": ATLAS_PATH,
    "atlas_image_format": "{0}/IMAGES/CT.nii.gz",
    "atlas_label_format": "{0}/STRUCTURES/{1}.nii.gz",
    "crop_atlas_to_structures": True,
    "crop_atlas_expansion_mm": (50, 50, 50),
    "guide_structure_name": "Heart",
    "superior_extension": 30,
}

OPEN_ATLAS_SETTINGS["label_fusion_settings"] = {
    "vote_type": "unweighted",
    "vote_params": None,
    "optimal_threshold": {
        "Atrium_L": 0.5,
        "Ventricle_L": 0.5,
        "Atrium_R": 0.5,
        "Ventricle_R": 0.5,
        "A_Aorta": 0.5,
        "A_Pulmonary": 0.5,
        "V_Venacava_S": 0.5,
        "Heart": 0.5,
    },
}

OPEN_ATLAS_SETTINGS["vessel_spline_settings"] = {
    "vessel_name_list": [
        "A_LAD",
        "A_Cflx",
        "A_Coronary_L",
        "A_Coronary_R",
    ],
    "vessel_radius_mm_dict": {
        "A_LAD": 2,
        "A_Cflx": 2,
        "A_Coronary_L": 2,
        "A_Coronary_R": 2,
    },
    "scan_direction_dict": {
        "A_LAD": "z",
        "A_Cflx": "z",
        "A_Coronary_L": "x",
        "A_Coronary_R": "z",
    },
    "stop_condition_type_dict": {
        "A_LAD": "count",
        "A_Cflx": "count",
        "A_Coronary_L": "count",
        "A_Coronary_R": "count",
    },
    "stop_condition_value_dict": {
        "A_LAD": 2,
        "A_Cflx": 2,
        "A_Coronary_L": 2,
        "A_Coronary_R": 2,
    },
}

OPEN_ATLAS_SETTINGS["geometric_segmentation_settings"]["atlas_structure_names"] = {
    "atlas_left_ventricle": "Ventricle_L",
    "atlas_right_ventricle": "Ventricle_R",
    "atlas_left_atrium": "Atrium_L",
    "atlas_right_atrium": "Atrium_R",
    "atlas_ascending_aorta": "A_Aorta",
    "atlas_pulmonary_artery": "A_Pulmonary",
    "atlas_superior_vena_cava": "V_Venacava_S",
    "atlas_whole_heart": "Heart",
}

OPEN_ATLAS_SETTINGS["postprocessing_settings"]["structures_for_binaryfillhole"] = [
    "Atrium_L",
    "Ventricle_L",
    "Atrium_R",
    "Ventricle_R",
    "A_Aorta",
    "A_Pulmonary",
    "V_Venacava_S",
    "Heart",
]

OPEN_ATLAS_SETTINGS["postprocessing_settings"]["structures_for_overlap_correction"] = [
    "Atrium_L",
    "Ventricle_L",
    "Atrium_R",
    "Ventricle_R",
    "A_Aorta",
    "A_Pulmonary",
    "V_Venacava_S",
]

OPEN_ATLAS_SETTINGS["return_proba_as_contours"] = True

HYBRID_SETTINGS_DEFAULTS = {
    "fetch_open_atlas": True,
    "nnunet_settings": NNUNET_SETTINGS_DEFAULTS,
    "cardiac_settings": OPEN_ATLAS_SETTINGS,
}
HYBRID_SETTINGS_DEFAULTS["nnunet_settings"]["folds"] = "all"


def install_open_atlas(atlas_path):
    """Fetch atlas from Zenodo and place into atlas_path

    Args:
        atlas_path (pathlib.Path): Path in which to place the atlas
    """

    logger.info("Fetching and installing open cardiac atlas to %s", atlas_path)
    temp_dir = tempfile.mkdtemp()
    download_and_extract_zip_file(OPEN_ATLAS_URL, temp_dir)
    temp_atlas_path = Path(temp_dir).joinpath("test_atlas")
    if not atlas_path.parent.exists():
        atlas_path.parent.mkdir(parents=True)
    shutil.copytree(temp_atlas_path, atlas_path)
    shutil.rmtree(temp_dir)


def install_atlas_from_zipfile(zip_file_path, atlas_path):

    with tempfile.TemporaryDirectory() as temp_dir:

        temp_atlas_path = Path(temp_dir).joinpath("test_atlas")

        with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
            zip_ref.extractall(temp_atlas_path)

        if not atlas_path.parent.exists():
            atlas_path.parent.mkdir(parents=True)

        shutil.copytree(temp_atlas_path, atlas_path)


def display_open_cardiac_zip_url_locations():
    open_nnunet_heart_model = "Task400_OPEN_HEART_1FOLD"

    nnunet_models = available_nnunet_models()
    nnunet_zip_url = nnunet_models[open_nnunet_heart_model]["url"]

    print("Please download the following two zip files:")
    print(nnunet_zip_url)
    print(OPEN_ATLAS_URL)
    print()
    print(
        "Once downloaded, pass where these are located on your filesystem to the "
        "install_hybrid_cardiac_from_zip function."
    )


def install_hybrid_cardiac_from_zip(path_to_nnunet_zip, path_to_atlas_zip):

    from nnunet.inference.pretrained_models.download_pretrained_model import (
        install_model_from_zip_file,
    )

    # Install nnUNet model
    setup_nnunet_environment()
    install_model_from_zip_file(path_to_nnunet_zip)

    # Install atlas model
    install_atlas_from_zipfile(path_to_atlas_zip, Path(ATLAS_PATH).parent)


def run_hybrid_segmentation(img, settings=HYBRID_SETTINGS_DEFAULTS):
    """Runs the hybrid cardiac segmentation

    Args:
        img (sitk.Image):
        settings (dict, optional): Dictionary containing settings for algorithm.
                                   Defaults to HYBRID_SETTINGS_DEFAULTS.

    Returns:
        dict: Dictionary containing output of segmentation
    """

    # Make sure atlas path exists, if not fetch it if fetch open atlas setting is true
    atlas_path = Path(settings["cardiac_settings"]["atlas_settings"]["atlas_path"])
    if not atlas_path.exists() or len(list(atlas_path.glob("*"))) == 0:
        if settings["fetch_open_atlas"]:
            # Fetch data from Zenodo
            install_open_atlas(atlas_path)
        else:
            raise SystemError(f"No atlas exists at {atlas_path}")

    # Run the whole heart nnUNet segmentation
    mask_wh = run_segmentation(img, settings["nnunet_settings"])

    # Run the 2nd part of the hybrid approach
    return run_cardiac_segmentation(
        img, guide_structure=mask_wh["Struct_0"], settings=settings["cardiac_settings"]
    )


def run_cardiac_segmentation(img, guide_structure=None, settings=CARDIAC_SETTINGS_DEFAULTS):
    """Runs the atlas-based cardiac segmentation

    Args:
        img (sitk.Image):
        settings (dict, optional): Dictionary containing settings for algorithm.
                                   Defaults to default_settings.

    Returns:
        dict: Dictionary containing output of segmentation
    """

    results = {}
    results_prob = {}

    return_as_cropped = settings["return_as_cropped"]

    """
    Initialisation - Read in atlases
    - image files
    - structure files

        Atlas structure:
        'ID': 'Original': 'CT Image'    : sitk.Image
                            'Struct A'    : sitk.Image
                            'Struct B'    : sitk.Image
                'RIR'     : 'CT Image'    : sitk.Image
                            'Transform'   : transform parameter map
                            'Struct A'    : sitk.Image
                            'Struct B'    : sitk.Image
                'DIR'     : 'CT Image'    : sitk.Image
                            'Transform'   : displacement field transform
                            'Weight Map'  : sitk.Image
                            'Struct A'    : sitk.Image
                            'Struct B'    : sitk.Image


    """

    logger.info("")
    # Settings
    atlas_path = settings["atlas_settings"]["atlas_path"]

    atlas_id_list = settings["atlas_settings"]["atlas_id_list"]
    atlas_structure_list = settings["atlas_settings"]["atlas_structure_list"]

    atlas_image_format = settings["atlas_settings"]["atlas_image_format"]
    atlas_label_format = settings["atlas_settings"]["atlas_label_format"]

    crop_atlas_to_structures = settings["atlas_settings"]["crop_atlas_to_structures"]
    crop_atlas_expansion_mm = settings["atlas_settings"]["crop_atlas_expansion_mm"]

    atlas_set = {}
    for atlas_id in atlas_id_list:
        atlas_set[atlas_id] = {}
        atlas_set[atlas_id]["Original"] = {}

        image = sitk.ReadImage(f"{atlas_path}/{atlas_image_format.format(atlas_id)}")

        structures = {
            struct: sitk.ReadImage(f"{atlas_path}/{atlas_label_format.format(atlas_id, struct)}")
            for struct in atlas_structure_list
        }

        if crop_atlas_to_structures:
            logger.info("Automatically cropping atlas: %s", atlas_id)

            original_volume = np.product(image.GetSize())

            crop_box_size, crop_box_index = label_to_roi(
                structures.values(), expansion_mm=crop_atlas_expansion_mm
            )

            image = crop_to_roi(image, size=crop_box_size, index=crop_box_index)

            final_volume = np.product(image.GetSize())

            logger.info("  > Volume reduced by factor %.2f", original_volume / final_volume)

            for struct in atlas_structure_list:
                structures[struct] = crop_to_roi(
                    structures[struct], size=crop_box_size, index=crop_box_index
                )

        atlas_set[atlas_id]["Original"]["CT Image"] = image

        for struct in atlas_structure_list:
            atlas_set[atlas_id]["Original"][struct] = structures[struct]

    """
    Step 1 - Automatic cropping
    If we have a guide structure:
        - use structure to crop target image

    Otherwise:
        - using a quick registration to register each atlas
        - expansion of the bounding box to ensure entire volume of interest is enclosed
        - target image is cropped
    """

    expansion_mm = settings["auto_crop_target_image_settings"]["expansion_mm"]

    if guide_structure:

        crop_box_size, crop_box_index = label_to_roi(guide_structure, expansion_mm=expansion_mm)
        img_crop = crop_to_roi(img, crop_box_size, crop_box_index)

        guide_structure = crop_to_roi(guide_structure, crop_box_size, crop_box_index)
        target_reg_structure = convert_mask_to_reg_structure(guide_structure, expansion=2)

    else:
        quick_reg_settings = {
            "reg_method": "similarity",
            "shrink_factors": [8],
            "smooth_sigmas": [0],
            "sampling_rate": 0.75,
            "default_value": -1000,
            "number_of_iterations": 25,
            "final_interp": sitk.sitkLinear,
            "metric": "mean_squares",
            "optimiser": "gradient_descent_line_search",
        }

        registered_crop_images = []

        logger.info("Running initial Translation tranform to crop image volume")

        for atlas_id in atlas_id_list[: min([8, len(atlas_id_list)])]:

            logger.info("  > atlas %s", atlas_id)

            # Register the atlases
            atlas_set[atlas_id]["RIR"] = {}
            atlas_image = atlas_set[atlas_id]["Original"]["CT Image"]

            reg_image, _ = linear_registration(
                img,
                atlas_image,
                **quick_reg_settings,
            )

            registered_crop_images.append(sitk.Cast(reg_image, sitk.sitkFloat32))

            del reg_image

        combined_image = sum(registered_crop_images) / len(registered_crop_images) > -1000

        crop_box_size, crop_box_index = label_to_roi(combined_image, expansion_mm=expansion_mm)

        img_crop = crop_to_roi(img, crop_box_size, crop_box_index)

    logger.info("Calculated crop box:")
    logger.info("  > %s", crop_box_index)
    logger.info("  > %s", crop_box_size)
    logger.info("  > Vol reduction = %.2f", np.product(img.GetSize()) / np.product(crop_box_size))

    """
    Step 2 - Rigid registration of target images
    - Individual atlas images are registered to the target
    - The transformation is used to propagate the labels onto the target
    """
    linear_registration_settings = settings["linear_registration_settings"]

    logger.info(
        "Running %s tranform to align atlas images", linear_registration_settings["reg_method"]
    )

    for atlas_id in atlas_id_list:
        # Register the atlases

        logger.info("  > atlas %s", atlas_id)

        atlas_set[atlas_id]["RIR"] = {}

        if guide_structure:
            guide_structure_name = settings["atlas_settings"]["guide_structure_name"]
            target_reg_image = target_reg_structure
            atlas_reg_image = convert_mask_to_reg_structure(
                atlas_set[atlas_id]["Original"][guide_structure_name], expansion=2
            )

        else:
            target_reg_image = img_crop
            atlas_reg_image = atlas_set[atlas_id]["Original"]["CT Image"]

        _, initial_tfm = linear_registration(
            target_reg_image,
            atlas_reg_image,
            **linear_registration_settings,
        )

        # Save in the atlas dict
        atlas_set[atlas_id]["RIR"]["Transform"] = initial_tfm

        if guide_structure:
            atlas_set[atlas_id]["RIR"]["Reg Mask"] = apply_transform(
                input_image=atlas_reg_image,
                reference_image=img_crop,
                transform=initial_tfm,
                default_value=0,
                interpolator=sitk.sitkLinear,
            )

            expanded_atlas_guide_structure = extend_mask(
                atlas_set[atlas_id]["Original"][guide_structure_name],
                direction=("ax", "sup"),
                extension_mm=settings["atlas_settings"]["superior_extension"],
                interior_mm_shape=settings["atlas_settings"]["superior_extension"] / 2,
            )

            atlas_set[atlas_id]["RIR"][guide_structure_name + "EXPANDED"] = apply_transform(
                input_image=expanded_atlas_guide_structure,
                reference_image=img_crop,
                transform=initial_tfm,
                default_value=0,
                interpolator=sitk.sitkNearestNeighbor,
            )

        atlas_set[atlas_id]["RIR"]["CT Image"] = apply_transform(
            input_image=atlas_set[atlas_id]["Original"]["CT Image"],
            reference_image=img_crop,
            transform=initial_tfm,
            default_value=-1000,
            interpolator=sitk.sitkLinear,
        )

        # sitk.WriteImage(rigid_image, f"./RR_{atlas_id}.nii.gz")

        for struct in atlas_structure_list:
            input_struct = atlas_set[atlas_id]["Original"][struct]
            atlas_set[atlas_id]["RIR"][struct] = apply_transform(
                input_image=input_struct,
                reference_image=img_crop,
                transform=initial_tfm,
                default_value=0,
                interpolator=sitk.sitkNearestNeighbor,
            )

        atlas_set[atlas_id]["Original"] = None

    """
    Step 3 - Deformable image registration
    - Using Fast Symmetric Diffeomorphic Demons
    """
    if guide_structure:
        structure_guided_registration_settings = settings["structure_guided_registration_settings"]

        logger.info("Running structure-guided deformable registration on atlas labels")

        for atlas_id in atlas_id_list:

            logger.info("  > atlas %s", atlas_id)

            # Register the atlases
            atlas_set[atlas_id]["DIR_STRUCT"] = {}

            deform_image, struct_guided_tfm, _ = fast_symmetric_forces_demons_registration(
                target_reg_structure,
                atlas_set[atlas_id]["RIR"]["Reg Mask"],
                **structure_guided_registration_settings,
            )

            # Save in the atlas dict
            atlas_set[atlas_id]["DIR_STRUCT"]["Reg Mask"] = deform_image
            atlas_set[atlas_id]["DIR_STRUCT"]["Transform"] = struct_guided_tfm

            atlas_set[atlas_id]["DIR_STRUCT"]["CT Image"] = apply_transform(
                input_image=atlas_set[atlas_id]["RIR"]["CT Image"],
                transform=struct_guided_tfm,
                default_value=-1000,
                interpolator=sitk.sitkLinear,
            )

            atlas_set[atlas_id]["DIR_STRUCT"][guide_structure_name + "EXPANDED"] = apply_transform(
                input_image=atlas_set[atlas_id]["RIR"][guide_structure_name + "EXPANDED"],
                reference_image=img_crop,
                transform=struct_guided_tfm,
                default_value=0,
                interpolator=sitk.sitkNearestNeighbor,
            )

            # sitk.WriteImage(deform_image, f"./DIR_STRUCT_{atlas_id}.nii.gz")

            for struct in atlas_structure_list:
                input_struct = atlas_set[atlas_id]["RIR"][struct]
                atlas_set[atlas_id]["DIR_STRUCT"][struct] = apply_transform(
                    input_image=input_struct,
                    transform=struct_guided_tfm,
                    default_value=0,
                    interpolator=sitk.sitkNearestNeighbor,
                )

            atlas_set[atlas_id]["RIR"] = None

    # Settings
    deformable_registration_settings = settings["deformable_registration_settings"]

    logger.info("Running DIR to refine atlas image registration")

    for atlas_id in atlas_id_list:

        logger.info("  > atlas %s", atlas_id)

        # Register the atlases
        atlas_set[atlas_id]["DIR"] = {}

        if guide_structure:
            label = "DIR_STRUCT"
        else:
            label = "RIR"

        atlas_reg_image = atlas_set[atlas_id][label]["CT Image"]
        target_reg_image = img_crop

        if guide_structure:
            expanded_atlas_mask = atlas_set[atlas_id]["DIR_STRUCT"][
                guide_structure_name + "EXPANDED"
            ]
            expanded_target_mask = extend_mask(
                guide_structure,
                direction=("ax", "sup"),
                extension_mm=settings["atlas_settings"]["superior_extension"],
                interior_mm_shape=settings["atlas_settings"]["superior_extension"] / 2,
            )

            combined_mask = sitk.Maximum(expanded_atlas_mask, expanded_target_mask)

            atlas_reg_image = sitk.Mask(atlas_reg_image, combined_mask, outsideValue=-1000)
            atlas_reg_image = sitk.Mask(
                atlas_reg_image, atlas_reg_image > -400, outsideValue=-1000
            )

            target_reg_image = sitk.Mask(target_reg_image, combined_mask, outsideValue=-1000)
            target_reg_image = sitk.Mask(
                target_reg_image, atlas_reg_image > -400, outsideValue=-1000
            )

        deform_image, dir_tfm, _ = fast_symmetric_forces_demons_registration(
            target_reg_image,
            atlas_reg_image,
            **deformable_registration_settings,
        )

        # Save in the atlas dict
        atlas_set[atlas_id]["DIR"]["Transform"] = dir_tfm

        atlas_set[atlas_id]["DIR"]["CT Image"] = apply_transform(
            input_image=atlas_set[atlas_id][label]["CT Image"],
            transform=dir_tfm,
            default_value=-1000,
            interpolator=sitk.sitkLinear,
        )

        for struct in atlas_structure_list:
            input_struct = atlas_set[atlas_id][label][struct]
            atlas_set[atlas_id]["DIR"][struct] = apply_transform(
                input_image=input_struct,
                transform=dir_tfm,
                default_value=0,
                interpolator=sitk.sitkNearestNeighbor,
            )

        atlas_set[atlas_id][label] = None

    """
    Step 4 - Iterative atlas removal
    - This is an automatic process that will attempt to remove inconsistent atlases from the entire set

    """
    # Compute weight maps
    # Here we use simple GWV as this minises the potentially negative influence of mis-registered
    # atlases
    iar_settings = settings["iar_settings"]

    if iar_settings["reference_structure"]:

        for atlas_id in atlas_id_list:
            atlas_image = atlas_set[atlas_id]["DIR"]["CT Image"]
            weight_map = compute_weight_map(img_crop, atlas_image, vote_type="global")
            atlas_set[atlas_id]["DIR"]["Weight Map"] = weight_map

        atlas_set = run_iar(atlas_set=atlas_set, **iar_settings)

    else:
        logger.info("IAR: No reference structure, skipping iterative atlas removal.")

    """
    Step 4 - Vessel Splining

    """
    vessel_spline_settings = settings["vessel_spline_settings"]

    if len(vessel_spline_settings["vessel_name_list"]) > 0:

        segmented_vessel_dict = vessel_spline_generation(
            img_crop, atlas_set, **vessel_spline_settings
        )
    else:
        logger.info("No vessel splining required, continue.")

    """
    Step 5 - Label Fusion
    """
    # Compute weight maps
    vote_type = settings["label_fusion_settings"]["vote_type"]
    vote_params = settings["label_fusion_settings"]["vote_params"]

    # Compute weight maps
    for atlas_id in list(atlas_set.keys()):
        atlas_image = atlas_set[atlas_id]["DIR"]["CT Image"]
        weight_map = compute_weight_map(
            img_crop, atlas_image, vote_type=vote_type, vote_params=vote_params
        )
        atlas_set[atlas_id]["DIR"]["Weight Map"] = weight_map

    combined_label_dict = combine_labels(atlas_set, atlas_structure_list)

    """
    Step 6 - Paste the cropped structure into the original image space
    """
    logger.info("Generating binary segmentations.")
    template_img_binary = sitk.Cast((img * 0), sitk.sitkUInt8)
    template_img_prob = sitk.Cast((img * 0), sitk.sitkFloat64)

    vote_structures = settings["label_fusion_settings"]["optimal_threshold"].keys()
    vote_structures = [i for i in vote_structures if i in atlas_structure_list]

    for structure_name in vote_structures:

        probability_map = combined_label_dict[structure_name]

        optimal_threshold = settings["label_fusion_settings"]["optimal_threshold"][structure_name]

        binary_struct = process_probability_image(probability_map, optimal_threshold)

        if return_as_cropped:
            results[structure_name] = binary_struct

            if settings["return_proba_as_contours"]:
                atlas_contours = [
                    process_probability_image(atlas_set[atlas_id]["DIR"][structure_name], 0.5)
                    for atlas_id in atlas_id_list
                ]

                results_prob[structure_name] = binary_encode_structure_list(atlas_contours)

            else:
                results_prob[structure_name] = probability_map

            # We also generate another version of the guide_structure using the atlas contours
            # We *can* return this, but probably don't want to
            # Here this check is performed
            if (not settings["return_atlas_guide_structure"]) and (guide_structure is not None):
                results[guide_structure_name] = guide_structure
                results_prob[guide_structure_name] = guide_structure

        else:

            if settings["return_proba_as_contours"]:
                atlas_contours = [
                    process_probability_image(atlas_set[atlas_id]["DIR"][structure_name], 0.5)
                    for atlas_id in atlas_id_list
                ]
                probability_img = binary_encode_structure_list(atlas_contours)
                template_img_prob = sitk.Cast((img * 0), sitk.sitkUInt32)

            else:
                probability_img = probability_map

            # Un-crop binary structure
            paste_img_binary = sitk.Paste(
                template_img_binary,
                binary_struct,
                binary_struct.GetSize(),
                (0, 0, 0),
                crop_box_index,
            )
            results[structure_name] = paste_img_binary

            # Un-crop probability map
            paste_prob_img = sitk.Paste(
                template_img_prob,
                probability_img,
                probability_img.GetSize(),
                (0, 0, 0),
                crop_box_index,
            )
            results_prob[structure_name] = paste_prob_img

            # Un-crop the guide structure
            if (not settings["return_atlas_guide_structure"]) and (guide_structure is not None):
                new_guide_structure = sitk.Paste(
                    template_img_binary,
                    guide_structure,
                    guide_structure.GetSize(),
                    (0, 0, 0),
                    crop_box_index,
                )
                results[guide_structure_name] = new_guide_structure
                results_prob[guide_structure_name] = new_guide_structure

    for structure_name in vessel_spline_settings["vessel_name_list"]:
        binary_struct = segmented_vessel_dict[structure_name]

        if return_as_cropped:
            results[structure_name] = binary_struct

            vessel_list = [
                atlas_set[atlas_id]["DIR"][structure_name] for atlas_id in list(atlas_set.keys())
            ]

        else:
            # Un-crop binary vessel
            paste_img_binary = sitk.Paste(
                template_img_binary,
                binary_struct,
                binary_struct.GetSize(),
                (0, 0, 0),
                crop_box_index,
            )
            results[structure_name] = paste_img_binary

            vessel_list = []
            for atlas_id in list(atlas_set.keys()):
                paste_img_binary = sitk.Paste(
                    template_img_binary,
                    atlas_set[atlas_id]["DIR"][structure_name],
                    atlas_set[atlas_id]["DIR"][structure_name].GetSize(),
                    (0, 0, 0),
                    crop_box_index,
                )
                vessel_list.append(paste_img_binary)

        # Encode list of vessels
        encoded_vessels = binary_encode_structure_list(vessel_list)
        results_prob[structure_name] = encoded_vessels

    """
    Step 7 - Geometric definitions of cardiac valves and conduction system nodes
    """
    geometric_segmentation_settings = settings["geometric_segmentation_settings"]

    if geometric_segmentation_settings["run_geometric_algorithms"]:

        logger.info("Computing geometric definitions for valves and conduction system.")

        geom_atlas_names = geometric_segmentation_settings["atlas_structure_names"]
        geom_valve_defs = geometric_segmentation_settings["valve_definitions"]
        geom_conduction_defs = geometric_segmentation_settings["conduction_system_definitions"]

        # 1 - MITRAL VALVE
        mv_name = "Valve_Mitral"
        results[mv_name] = generate_valve_using_cylinder(
            label_atrium=results[geom_atlas_names["atlas_left_atrium"]],
            label_ventricle=results[geom_atlas_names["atlas_left_ventricle"]],
            radius_mm=geom_valve_defs["mitral_valve_radius_mm"],
            height_mm=geom_valve_defs["mitral_valve_thickness_mm"],
        )

        # 2 - TRICUSPID VALVE
        tv_name = "Valve_Tricuspid"
        results[tv_name] = generate_valve_using_cylinder(
            label_atrium=results[geom_atlas_names["atlas_right_atrium"]],
            label_ventricle=results[geom_atlas_names["atlas_right_ventricle"]],
            radius_mm=geom_valve_defs["tricuspid_valve_radius_mm"],
            height_mm=geom_valve_defs["tricuspid_valve_thickness_mm"],
        )

        # 3 - AORTIC VALVE
        av_name = "Valve_Aortic"
        results[av_name] = generate_valve_from_great_vessel(
            label_great_vessel=results[geom_atlas_names["atlas_ascending_aorta"]],
            label_ventricle=results[geom_atlas_names["atlas_left_ventricle"]],
            valve_thickness_mm=geom_valve_defs["aortic_valve_thickness_mm"],
        )

        # 4 - PULMONIC VALVE
        pv_name = "Valve_Pulmonic"
        results[pv_name] = generate_valve_from_great_vessel(
            label_great_vessel=results[geom_atlas_names["atlas_pulmonary_artery"]],
            label_ventricle=results[geom_atlas_names["atlas_right_ventricle"]],
            valve_thickness_mm=geom_valve_defs["pulmonic_valve_thickness_mm"],
        )

        # 5 - SINOATRIAL NODE
        san_name = "CN_Sinoatrial"
        results[san_name] = geometric_sinoatrialnode(
            label_svc=results[geom_atlas_names["atlas_superior_vena_cava"]],
            label_ra=results[geom_atlas_names["atlas_right_atrium"]],
            label_wholeheart=results[geom_atlas_names["atlas_whole_heart"]],
            radius_mm=geom_conduction_defs["sinoatrial_node_radius_mm"],
        )

        # 6 - ATRIOVENTRICULAR NODE
        avn_name = "CN_Atrioventricular"
        results[avn_name] = geometric_atrioventricularnode(
            label_la=results[geom_atlas_names["atlas_left_atrium"]],
            label_lv=results[geom_atlas_names["atlas_left_ventricle"]],
            label_ra=results[geom_atlas_names["atlas_right_atrium"]],
            label_rv=results[geom_atlas_names["atlas_right_ventricle"]],
            radius_mm=geom_conduction_defs["atrioventricular_node_radius_mm"],
        )

    """
    Step 8 - Post-processing
    """
    postprocessing_settings = settings["postprocessing_settings"]

    if postprocessing_settings["run_postprocessing"]:
        logger.info("Running post-processing.")

        # Remove any smaller components and perform morphological closing (hole filling)
        binaryfillhole_img = [
            int(postprocessing_settings["binaryfillhole_mm"] / sp) for sp in img.GetSpacing()
        ]

        for structure_name in postprocessing_settings["structures_for_binaryfillhole"]:

            if structure_name not in results.keys():
                continue

            contour_s = results[structure_name]
            contour_s = sitk.RelabelComponent(sitk.ConnectedComponent(contour_s)) == 1
            contour_s = sitk.BinaryMorphologicalClosing(contour_s, binaryfillhole_img)
            results[structure_name] = contour_s

        # Remove any overlaps
        input_overlap = {
            s: results[s] for s in postprocessing_settings["structures_for_overlap_correction"]
        }
        output_overlap = correct_volume_overlap(input_overlap)

        for s in postprocessing_settings["structures_for_overlap_correction"]:
            results[s] = output_overlap[s]

    if return_as_cropped:
        results["CROP_IMAGE"] = img_crop

    logger.info("Done!")

    return results, results_prob



__project__ = "platipy"
__version__ = "0.6.1"
__keywords__ = [
    "medical imaging",
    "visualisation",
    "registration",
    "radiotherapy",
    "image analysis",
]
__author__ = "Phillip Chlap & Robert Finnegan"
__author_email__ = "phillip.chlap@unsw.edu.au"
__url__ = "https://pyplati.github.io/platipy/"
__platforms__ = "ALL"
